@@ -2,7 +2,7 @@
 ** Modular Logfile Analyzer
 ** Copyright 2000 Jan Kneschke <jan@kneschke.de>
 **
-** Homepage: http://www.kneschke.de/projekte/modlogan
+** Homepage: http://www.modlogan.org
 **
 
     This program is free software; you can redistribute it and/or modify
@@ -22,7 +22,7 @@
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 
 **
-** $Id: parse.c,v 1.7 2001/12/28 17:57:26 ostborn Exp $
+** $Id: parse.c,v 1.13 2003/04/18 18:37:45 ostborn Exp $
 */
 
 #include <stdlib.h>
@@ -34,10 +34,6 @@
 
 #include "config.h"
 
-#ifdef HAVE_LIBADNS
-#include <adns.h>
-#endif
-
 #include "mlocale.h"
 #include "mplugins.h"
 #include "mrecord.h"
@@ -51,40 +47,43 @@
 int find_os (mconfig *ext_conf, char *str) {
 	config_input *conf = ext_conf->plugin_conf;
 	mlist *l = conf->match_os;
+	int str_len;
 	if (!str || !l) return 0;
-	
+
 	while (*str == ' ') str++;
 	
-	while (l) {
+	str_len = strlen(str);
+
+	for (l = conf->match_os; l; l = l->next) {
 		mdata *data = l->data;
-		
-		if (data && strmatch(data->data.match.match, NULL, str)) {
+
+		if (data && strmatch(data->data.match.match, NULL, str, str_len)) {
 			return 1;
 		}
-		
-		l = l->next;
 	}
-	
+
 	return 0;
 }
 
 int find_ua (mconfig *ext_conf, char *str) {
 	config_input *conf = ext_conf->plugin_conf;
 	mlist *l = conf->match_ua;
-	if (!str || !l) return 0;
+	int str_len;
 	
+	if (!str || !l) return 0;
+
 	while (*str == ' ') str++;
 	
-	while (l) {
+	str_len = strlen(str);
+
+	for (l = conf->match_ua; l; l = l->next) {
 		mdata *data = l->data;
-		
-		if (data && strmatch(data->data.match.match, NULL, str)) {
+
+		if (data && strmatch(data->data.match.match, NULL, str, str_len)) {
 			return 1;
 		}
-		
-		l = l->next;
 	}
-	
+
 	return 0;
 }
 
@@ -94,14 +93,14 @@ int parse_timestamp(mconfig *ext_conf, const char *_date, const char *_time, mlo
 	char buf[10];
 	struct tm tm;
 	config_input *conf = ext_conf->plugin_conf;
-	
+
 	char *str = NULL;
-	
+
 	str = malloc(strlen(_date) + strlen(_time) + 2);
 	strcpy(str, _date);
 	strcat(str, " ");
 	strcat(str, _time);
-	
+
 	if ((n = pcre_exec(conf->match_timestamp, conf->match_timestamp_extra, str, strlen(str), 0, 0, ovector, 3 * N)) < 0) {
 		if (n == PCRE_ERROR_NOMATCH) {
 			fprintf(stderr, "%s.%d: string doesn't match: %s\n", __FILE__, __LINE__, str);
@@ -110,23 +109,25 @@ int parse_timestamp(mconfig *ext_conf, const char *_date, const char *_time, mlo
 		}
 		return -1;
 	}
-	
+
+	memset(&tm, 0, sizeof(struct tm));
+
 	pcre_copy_substring(str, ovector, n, 1, buf, sizeof(buf));
 	tm.tm_year = strtol(buf, NULL, 10)-1900;
 	pcre_copy_substring(str, ovector, n, 3, buf, sizeof(buf));
 	tm.tm_mday = strtol(buf, NULL, 10);
 	pcre_copy_substring(str, ovector, n, 2, buf, sizeof(buf));
 	tm.tm_mon = strtol(buf, NULL, 10)-1;
-	
+
 	pcre_copy_substring(str, ovector, n, 4, buf, sizeof(buf));
 	tm.tm_hour = strtol(buf, NULL, 10);
 	pcre_copy_substring(str, ovector, n, 5, buf, sizeof(buf));
 	tm.tm_min = strtol(buf, NULL, 10);
 	pcre_copy_substring(str, ovector, n, 6, buf, sizeof(buf));
 	tm.tm_sec = strtol(buf, NULL, 10);
-	
+
 	record->timestamp = mktime (&tm);
-	
+
 	return 0;
 #undef  N
 }
@@ -134,18 +135,18 @@ int parse_timestamp(mconfig *ext_conf, const char *_date, const char *_time, mlo
 int parse_useragent(mconfig *ext_conf,const char *str, mlogrec_web_extclf *record) {
 /* get user agent */
 	char *pc1 = (char *)str, *pc3, *pc2 = (char *)str, *buf_copy;
-	
+
 	buf_copy = malloc(strlen(str)+1);
 	strcpy(buf_copy, str);
-	
+
 	str = urlescape((char *)str);
-	
+
 	if ((pc3 = strchr(pc1, '(') )) {
 		if (strstr(pc3, "compatible")) {
 			int finished = 0;
-			
+
 			pc1 = pc2 = (pc3+1);
-			
+
 			while (!finished) {
 				while (*pc2 && !(*pc2 == ';' || *pc2 == ')')) pc2++;
 				if (!*pc2) {
@@ -156,35 +157,32 @@ int parse_useragent(mconfig *ext_conf,const char *str, mlogrec_web_extclf *recor
 				} else if (*pc2 == ')') {
 					finished = 1;
 				}
-				
+
 				while (*pc1 == '_') pc1++;
-				
+
 				*pc2 = '\0';
 				if (!record->req_useragent && find_ua(ext_conf, pc1)) {
-					record->req_useragent = malloc(pc2-pc1+1);
-					strcpy(record->req_useragent, pc1);
+					buffer_strcpy(record->req_useragent, pc1);
 				} else if (!record->req_useros && find_os(ext_conf, pc1)) {
-					record->req_useros = malloc(pc2-pc1+1);
-					strcpy(record->req_useros, pc1);
+					buffer_strcpy(record->req_useros, pc1);
 				}
 				pc1 = ++pc2;
 			}
-			
+
 
 		} else {
 			int finished = 0;
-			
+
 			pc2 = pc3;
 
 			*pc2 = '\0';
-#if 0			
+#if 0
 			if (!find_ua(ext_conf, pc1)) {
 				printf("UA- unknown: %s\n", pc4);
 			}
 #endif
-			record->req_useragent = malloc(pc2-pc1+1);
-			strcpy(record->req_useragent, pc1);
-			
+			buffer_strcpy(record->req_useragent, pc1);
+
 			pc1 = pc2 = (pc3+1);
 
 			while (!finished) {
@@ -199,13 +197,12 @@ int parse_useragent(mconfig *ext_conf,const char *str, mlogrec_web_extclf *recor
 				}
 
 				while (*pc1 == '_') pc1++;
-				
+
 				*pc2 = '\0';
-								
-				
+
+
 				if (!record->req_useros && find_os(ext_conf, pc1)) {
-					record->req_useros = malloc(strlen(pc1)+1);
-					strcpy(record->req_useros, pc1);
+					buffer_strcpy(record->req_useros, pc1);
 				}
 				pc1 = ++pc2;
 			}
@@ -215,18 +212,17 @@ int parse_useragent(mconfig *ext_conf,const char *str, mlogrec_web_extclf *recor
 		if (!record->req_useragent) {
 			printf("UA unknown: %s\n", pc4);
 		}
-		
+
 		if (!record->req_useros) {
 			printf("OS unknown: %s\n", pc4);
 		}
 #endif
 	} else {
-		record->req_useragent = malloc(strlen(str)+1);
-		strcpy(record->req_useragent, str);
+		buffer_strcpy(record->req_useragent, str);
 	}
-	
+
 	free(buf_copy);
-	
+
 	return 0;
 }
 
@@ -235,7 +231,7 @@ int parse_referrer(mconfig *ext_conf,const char *str, mlogrec_web_extclf *record
 	int ovector[3 * N], n;
 	config_input *conf = ext_conf->plugin_conf;
 	const char **list;
-	
+
 	if ((n = pcre_exec(conf->match_referrer, conf->match_referrer_extra, str, strlen(str), 0, 0, ovector, 3 * N)) < 0) {
 		if (n == PCRE_ERROR_NOMATCH) {
 			fprintf(stderr, "%s.%d: string doesn't match: %s\n", __FILE__, __LINE__, str);
@@ -244,19 +240,17 @@ int parse_referrer(mconfig *ext_conf,const char *str, mlogrec_web_extclf *record
 		}
 		return -1;
 	}
-	
+
 	if (n >= 2) {
 		/* everything has matched, take the different pieces and be happy :) */
 		pcre_get_substring_list(str, ovector, n, &list);
-	
-		record->ref_url = malloc(strlen((char *)list[1])+1);
-		strcpy(record->ref_url, (char *)list[1]);
-		
+
+		buffer_strcpy(record->ref_url, (char *)list[1]);
+
 		if (n > 3) {
-			record->ref_getvars = malloc(strlen((char *)list[3])+1);
-			strcpy(record->ref_getvars, (char *)list[3]);
+			buffer_strcpy(record->ref_getvars, (char *)list[3]);
 		}
-#ifdef DEBUG_INPUT		
+#ifdef DEBUG_INPUT
 		fprintf(stderr, "%s.%d: %s, %s\n", __FILE__, __LINE__, record->ref_url, record->ref_getvars);
 #endif
 		free(list);
@@ -264,9 +258,9 @@ int parse_referrer(mconfig *ext_conf,const char *str, mlogrec_web_extclf *record
 		fprintf(stderr, "%s.%d: Matched fields below minimum: %d\n", __FILE__, __LINE__, n);
 		return -1;
 	}
-	
-	
-	
+
+
+
 	return 0;
 #undef  N
 }
@@ -339,7 +333,7 @@ const msmedia_field_def def[] = {
 	{ "s-dns",	M_MSMEDIA_FIELD_SERVER_DNS,	"(.+?)"},
 	{ "s-totalclients",	M_MSMEDIA_FIELD_TOTALCLIENTS,	"(.+?)"},
 	{ "s-cpu-util",	M_MSMEDIA_FIELD_SERVER_CPU_UTIL,	"(.+?)"},
-	
+
 	{ NULL, 0, NULL}
 };
 
@@ -347,50 +341,43 @@ int parse_msmedia_field_info(mconfig *ext_conf, const char *_buffer) {
 	config_input *conf = ext_conf->plugin_conf;
 	char *buf, *pa, *pe;
 	int pos = 0, i;
-	char *match_buf;
+	buffer *match_buf;
 	const char *errptr;
 	int erroffset = 0;
-	
+
 	if (_buffer == NULL) {
 		fprintf(stderr, "%s.%d: buffer is empty\n", __FILE__, __LINE__);
 		return -1;
 	}
-	
-	if ((buf = malloc(strlen(_buffer)+1)) == NULL) {
-		fprintf(stderr, "%s.%d: malloc failed\n", __FILE__, __LINE__);
-		return -1;
-	}
-	
-	strcpy(buf, _buffer);
-	
+
+	buf = strdup(_buffer);
+
 	pa = buf;
 	pe = NULL;
-	
-	while ((pe = strchr(pa, ' ')) != NULL) {
+
+	for (pa = buf; (pe = strchr(pa, ' ')) != NULL; pa = pe + 1) {
 		*pe = '\0';
-		
+
 		for (i = 0; def[i].field != NULL; i++) {
 			if (strcmp(def[i].field, pa) == 0) {
 				break;
 			}
 		}
-		
+
 		if (def[i].field != NULL) {
 			if (pos >= M_MSMEDIA_MAX_FIELDS) {
 				fprintf(stderr, "%s.%d: not enough space to save the field defenition, increment M_MSMEDIA_MAX_FIELDS\n", __FILE__, __LINE__);
 				return -1;
 			}
-			
+
 			conf->trans_fields[pos++] = i;
 		} else {
 			fprintf(stderr, "%s.%d: Unknown fieldtype: %s\n", __FILE__, __LINE__, pa);
 			free(buf);
 			return -1;
 		}
-		
-		pa = pe + 1;
 	}
-	
+
 	/* don't forget the last param */
 	if (*pa) {
 		for (i = 0; def[i].field != NULL; i++) {
@@ -398,7 +385,7 @@ int parse_msmedia_field_info(mconfig *ext_conf, const char *_buffer) {
 				break;
 			}
 		}
-	
+
 		if (def[i].field != NULL) {
 			if (pos >= M_MSMEDIA_MAX_FIELDS) return -1;
 			conf->trans_fields[pos++] = i;
@@ -408,154 +395,101 @@ int parse_msmedia_field_info(mconfig *ext_conf, const char *_buffer) {
 			return -1;
 		}
 	}
-	
+
 	free(buf);
-	
-	match_buf = malloc(1024);
-	
-	*match_buf = '\0';
-	
+
+	match_buf = buffer_init();
+
 	for (i = 0; i < pos; i++) {
-		if (*match_buf == '\0') {
-			match_buf = strcat(match_buf, "^");
+		if (match_buf->used == 0) {
+			buffer_strcat(match_buf, "^");
 		} else {
-			match_buf = strcat(match_buf, " ");
+			buffer_strcat(match_buf, " ");
 		}
-		match_buf = strcat(match_buf, def[conf->trans_fields[i]].match);
+		buffer_strcat(match_buf, def[conf->trans_fields[i]].match);
 	}
-	
-	match_buf = strcat(match_buf, "$");
-	
-	if ((conf->match_clf = pcre_compile(match_buf,
+
+	buffer_strcat(match_buf, "$");
+
+	if ((conf->match_clf = pcre_compile(match_buf->ptr,
 		0, &errptr, &erroffset, NULL)) == NULL) {
-		
+
 		fprintf(stderr, "%s.%d: rexexp compilation error at %s\n", __FILE__, __LINE__, errptr);
-		free(match_buf);
+		buffer_free(match_buf);
 		return -1;
-	} 
-	free(match_buf);
-	
+	}
+	buffer_free(match_buf);
+
 	conf->match_clf_extra = pcre_study(conf->match_clf, 0, &errptr);
 	if (errptr != NULL) {
 		fprintf(stderr, "%s.%d: rexexp studying error at %s\n", __FILE__, __LINE__, errptr);
 		return -1;
 	}
-	
+
 	return 0;
 }
 
 int parse_msmedia_date_info(mconfig *ext_conf, const char *_buffer) {
-	/* getting 
+	/* getting
 		default_data
 		default_time
 	*/
 	config_input *conf = ext_conf->plugin_conf;
 	char *c;
-	
+
 	if ((c = strchr(_buffer, ' ')) != NULL) {
 		*c++ = '\0';
-		
-		conf->default_date = malloc(strlen(_buffer) + 1);
-		strcpy(conf->default_date, _buffer);
-		
-		conf->default_date = malloc(strlen(c) + 1);
-		strcpy(conf->default_date, c);
-		
+
+		conf->default_date = strdup(_buffer);
+		conf->default_time = strdup(c);
+
 		return 0;
 	} else {
 		return -1;
 	}
 }
 
-#ifdef HAVE_LIBADNS
-const char* reverse_ip(const char *ip) {
-	static char ip_buf[30];
-	pcre *m;
-#define N 20 + 1
-	const char **list;
-	int ovector[3 * N], n;
-	const char *errptr;
-	int erroffset = 0;
-	
-	if (!ip) return NULL;
-	
-	if ((m = pcre_compile(
-		"^([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3})\\.([0-9]{1,3})$", 
-		0, &errptr, &erroffset, NULL)) == NULL) {
-		
-		fprintf(stderr, "%s.%d: rexexp compilation error at %s\n", __FILE__, __LINE__, errptr);
-		return NULL;
-	} 
-	
-	if ((n = pcre_exec(m, NULL, ip, strlen(ip), 0, 0, ovector, 3 * N)) < 0) {
-		if (n != PCRE_ERROR_NOMATCH) {
-			fprintf(stderr, "%s.%d: execution error while matching: %d\n", __FILE__, __LINE__, n);
-		}
-		return NULL;
-	}
-	
-	pcre_get_substring_list(ip, ovector, n, &list);
-	
-	sprintf(ip_buf, "%s.%s.%s.%s.in-addr.arpa.",
-		list[4],
-		list[3],
-		list[2],
-		list[1]
-		);
-	
-	pcre_free(list);
-	free(m);
-#undef N
-	return ip_buf;
-}
-#endif
-
 /**
  * parse a msmedia record
- * 
+ *
  * @param ext_conf
  * @param record
  * @param _buffer
  * @returns 0  - no error, -1 - fatal error, 1  - corrupt record
  */
-int parse_record_pcre(mconfig *ext_conf, mlogrec *record, char *_buffer) {
+int parse_record_pcre(mconfig *ext_conf, mlogrec *record, buffer *b) {
 #define N M_MSMEDIA_MAX_FIELDS
 	const char **list;
 	int ovector[3 * N], n, i;
 	config_input *conf = ext_conf->plugin_conf;
-	mlogrec_web *recweb = NULL; 
-	mlogrec_web_extclf *recext = NULL; 
-	
+	mlogrec_web *recweb = NULL;
+	mlogrec_web_extclf *recext = NULL;
+
 	const char *_date = NULL, *_time = NULL;
-	
-#ifdef HAVE_LIBADNS
-	const char *conv_ip;
-	adns_query *query = NULL;
-	mlogrec_web *web = NULL;
-#endif
-	
+
 	/* remove the carriage return */
-	if (_buffer[strlen(_buffer)-1] == '\r') {
-		_buffer[strlen(_buffer)-1] = '\0';
+	if (b->ptr[b->used - 2] == '\r') {
+		b->ptr[b->used - 2] = '\0';
+		b->used --;
 	}
 
-	if (*_buffer == '#') { 
-		if (strncmp("#Version: ", _buffer, 10) == 0) {
-			if (strncmp("#Version: 4.1", _buffer, 13) != 0) {
+	if (*b->ptr == '#') {
+		if (strncmp("#Version: ", b->ptr, 10) == 0) {
+			if (strncmp("#Version: 4.1", b->ptr, 13) != 0) {
 				fprintf(stderr, "%s.%d: only logfile version 4.1 is supported\n", __FILE__, __LINE__);
 				return M_RECORD_HARD_ERROR;
 			} else {
 				return M_RECORD_IGNORED;
 			}
-		} else if (strncmp("#Fields: ", _buffer, 9) == 0) {
-			if (parse_msmedia_field_info(ext_conf, _buffer+9) != 0) {
+		} else if (strncmp("#Fields: ", b->ptr, 9) == 0) {
+			if (parse_msmedia_field_info(ext_conf, b->ptr+9) != 0) {
 				fprintf(stderr, "%s.%d: parsing #Fields: failed somehow\n", __FILE__, __LINE__);
 				return M_RECORD_HARD_ERROR;
 			} else {
 				return M_RECORD_IGNORED;
 			}
-		} else if (strncmp("#Date: ", _buffer, 7) == 0) {
-			if (parse_msmedia_date_info(ext_conf, _buffer+7) != 0) {
+		} else if (strncmp("#Date: ", b->ptr, 7) == 0) {
+			if (parse_msmedia_date_info(ext_conf, b->ptr+7) != 0) {
 				return M_RECORD_HARD_ERROR;
 			} else {
 				return M_RECORD_IGNORED;
@@ -564,24 +498,31 @@ int parse_record_pcre(mconfig *ext_conf, mlogrec *record, char *_buffer) {
 			return M_RECORD_IGNORED;
 		}
 	}
-	
+
 	if (conf->match_clf == NULL) return M_RECORD_HARD_ERROR;
 	
-	recweb = mrecord_init_web();
-	
-	record->ext_type = M_RECORD_TYPE_WEB;
-	record->ext = recweb;
-	
-	recext = mrecord_init_web_extclf();
-	
-	recweb->ext_type = M_RECORD_TYPE_WEB_EXTCLF;
-	recweb->ext = recext;
+	if (record->ext_type != M_RECORD_TYPE_WEB) {
+		if (record->ext_type != M_RECORD_TYPE_UNSET) {
+			mrecord_free_ext(record);
+		}
+		
+		record->ext_type = M_RECORD_TYPE_WEB;
+		record->ext = mrecord_init_web();
+	}
+
+	recweb = record->ext;
 	
 	if (recweb == NULL) return 1;
-/* parse a CLF record */	
-	if ((n = pcre_exec(conf->match_clf, conf->match_clf_extra, _buffer, strlen(_buffer), 0, 0, ovector, 3 * N)) < 0) {
+
+	recext = mrecord_init_web_extclf();
+
+	recweb->ext_type = M_RECORD_TYPE_WEB_EXTCLF;
+	recweb->ext = recext;
+
+/* parse a CLF record */
+	if ((n = pcre_exec(conf->match_clf, conf->match_clf_extra, b->ptr, b->used - 1, 0, 0, ovector, 3 * N)) < 0) {
 		if (n == PCRE_ERROR_NOMATCH) {
-			fprintf(stderr, "%s.%d: string doesn't match: %s\n", __FILE__, __LINE__, _buffer);
+			fprintf(stderr, "%s.%d: string doesn't match: %s\n", __FILE__, __LINE__, b->ptr);
 			return M_RECORD_CORRUPT;
 		} else {
 			fprintf(stderr, "%s.%d: execution error while matching: %d\n", __FILE__, __LINE__, n);
@@ -593,8 +534,8 @@ int parse_record_pcre(mconfig *ext_conf, mlogrec *record, char *_buffer) {
 		fprintf(stderr, "%s.%d: !REPORT ME! N is too low -> %d\n", __FILE__, __LINE__, N);
 		return M_RECORD_HARD_ERROR;
 	}
-	
-	pcre_get_substring_list(_buffer, ovector, n, &list);
+
+	pcre_get_substring_list(b->ptr, ovector, n, &list);
 
 	for (i = 0; i < n-1; i++) {
 		switch (def[conf->trans_fields[i]].id) {
@@ -605,36 +546,10 @@ int parse_record_pcre(mconfig *ext_conf, mlogrec *record, char *_buffer) {
 			_time = list[i+1];
 			break;
 		case M_MSMEDIA_FIELD_CLIENT_IP:
-			recweb->req_host = malloc(strlen((char *)list[i+1])+1);
-			strcpy(recweb->req_host, (char *)list[i+1]);
-#ifdef HAVE_LIBADNS
-			if (ext_conf->enable_resolver) {
-				conv_ip = reverse_ip(recweb->req_host);
-				
-				if (conv_ip) {
-					mdata *data = NULL;
-					
-					if (!mhash_in_hash(ext_conf->query_hash, recweb->req_host)) {
-						query = malloc(sizeof(adns_query));
-						adns_submit(*(ext_conf->adns),
-							    conv_ip,
-							    adns_r_ptr,
-							    adns_qf_quoteok_cname|adns_qf_cname_loose,
-							    NULL,
-							    query
-							    );
-						/* put ip and query into a hash */
-						data = mdata_Query_create(web->req_host, query);
-						mhash_insert_sorted(ext_conf->query_hash, data);
-					}
-					/* don't free query !! it will be removed by mhash_free */
-				}
-			}
-#endif
+			buffer_strcpy(recweb->req_host_ip, (char *)list[i+1]);
 			break;
 		case M_MSMEDIA_FIELD_URI_STEM:
-			recweb->req_url = malloc(strlen((char *)list[i+1])+1);
-			strcpy(recweb->req_url, (char *)list[i+1]);
+			buffer_strcpy(recweb->req_url, (char *)list[i+1]);
 			break;
 		case M_MSMEDIA_FIELD_CLIENT_STATUS:
 			recweb->req_status = strtol(list[i+1], NULL,10);
@@ -643,12 +558,10 @@ int parse_record_pcre(mconfig *ext_conf, mlogrec *record, char *_buffer) {
 			recweb->xfersize = strtod(list[i+1], NULL);
 			break;
 		case M_MSMEDIA_FIELD_SERVER_IP:
-			recext->srv_host = malloc(strlen((char *)list[i+1])+1);
-			strcpy(recext->srv_host, (char *)list[i+1]);
+			buffer_strcpy(recext->srv_host, (char *)list[i+1]);
 			break;
 		case M_MSMEDIA_FIELD_PROTOCOL:
-			recweb->req_protocol = malloc(strlen((char *)list[i+1])+1);
-			strcpy(recweb->req_protocol, (char *)list[i+1]);
+			buffer_strcpy(recweb->req_protocol, (char *)list[i+1]);
 			break;
 		case M_MSMEDIA_FIELD_USER_AGENT:
 			if (parse_useragent(ext_conf, list[i+1], recext)  == -1) {
@@ -700,7 +613,7 @@ int parse_record_pcre(mconfig *ext_conf, mlogrec *record, char *_buffer) {
 			/*if (ext_conf->debug_level > 2)
 				fprintf(stderr, "the field '%s' is known, but not supported yet.\n",def[conf->trans_fields[i]].field);*/
 			break;
-			
+
 
 		default:
 			fprintf(stderr, "the field '%s' is unknown\n", def[conf->trans_fields[i]].field);
@@ -711,7 +624,7 @@ int parse_record_pcre(mconfig *ext_conf, mlogrec *record, char *_buffer) {
 	if (_time != NULL && (_date != NULL || conf->default_date != NULL)) {
 		parse_timestamp(ext_conf, _date ? _date : conf->default_date, _time, record);
 	}
-	
+
 	free(list);
 
 	return M_RECORD_NO_ERROR;
@@ -719,25 +632,21 @@ int parse_record_pcre(mconfig *ext_conf, mlogrec *record, char *_buffer) {
 }
 
 int mplugins_input_msmedia_get_next_record(mconfig *ext_conf, mlogrec *record) {
+	int ret = 0;
 	config_input *conf = ext_conf->plugin_conf;
-	
-	if (record == NULL) return M_RECORD_HARD_ERROR;
-	
-	if (!fgets(conf->buffer, conf->buf_len-1,conf->inputfile)) {
-		return M_RECORD_EOF;
-	}
 
-	while (conf->buffer[strlen(conf->buffer)-1] != '\n') {
-		conf->buffer = realloc(conf->buffer, (conf->buf_len+conf->buf_inc) * sizeof(char));
-		
-		if (!fgets(conf->buffer+strlen(conf->buffer), conf->buf_inc-1,conf->inputfile)) {
-			return M_RECORD_EOF;
-		}
-		
-		conf->buf_len += conf->buf_inc;
+	if (record == NULL) return M_RECORD_HARD_ERROR;
+
+	/* fill the line buffer */
+	if (NULL == mgets(&(conf->inputfile), conf->buf)) return M_RECORD_EOF;
+	
+	ret = parse_record_pcre   (ext_conf, record, conf->buf);
+	
+	if (ret == M_RECORD_CORRUPT) {
+		M_DEBUG1(ext_conf->debug_level, M_DEBUG_SECTION_PARSING, M_DEBUG_LEVEL_WARNINGS,
+			 "affected Record: %s\n",
+			 conf->buf->ptr
+			 );
 	}
-	
-	conf->buffer[strlen(conf->buffer)-1] = '\0';
-	
-	return parse_record_pcre(ext_conf, record, conf->buffer);
+	return ret;
 }
