@@ -2,7 +2,7 @@
 ** Modular Logfile Analyzer
 ** Copyright 2000 Jan Kneschke <jan@kneschke.de>
 **
-** Homepage: http://www.kneschke.de/projekte/modlogan
+** Homepage: http://www.modlogan.org
 **
 
     This program is free software; you can redistribute it and/or modify
@@ -22,7 +22,7 @@
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 
 **
-** $Id: plugin_config.h,v 1.5 2001/06/04 22:11:15 jk Exp $
+** $Id: plugin_config.h,v 1.8 2003/04/18 18:40:45 ostborn Exp $
 */
 #ifndef _INPUT_PLUGIN_CONFIG_H_
 #define _INPUT_PLUGIN_CONFIG_H_
@@ -30,17 +30,13 @@
 #include <stdio.h>
 #include <pcre.h>
 
-#include "mlist.h"
-
+#include "mio.h"
 
 typedef struct {
-	FILE	*inputfile;
+	mfile	inputfile;
+	buffer	*buf;
 	char	*inputfilename;
-	
-	char	*buffer;
-	int	buf_len;
-	int	buf_inc;
-	
+
 	pcre	*match_clf;
 	pcre_extra	*match_clf_extra;
 	pcre	*match_timestamp;
