@@ -2,7 +2,7 @@
 ** Modular Logfile Analyzer
 ** Copyright 2000 Jan Kneschke <jan@kneschke.de>
 **
-** Homepage: http://www.kneschke.de/projekte/modlogan
+** Homepage: http://www.modlogan.org
 **
 
     This program is free software; you can redistribute it and/or modify
@@ -22,12 +22,12 @@
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 
 **
-** $Id: main.c,v 1.73 2002/01/03 12:13:08 ostborn Exp $
+** $Id: main.c,v 1.104 2003/04/18 18:14:29 ostborn Exp $
 */
 
 /**
  * \mainpage ModLogAn API Documenation
- * 
+ *
  * \section Introduction
  * Add docs.
  *
@@ -47,8 +47,6 @@
 #include <sys/time.h>
 #include <assert.h>
 
-#include <libxml/parser.h>
-#include <libxml/xmlerror.h>
 #include <zlib.h>
 
 #include "config.h"
@@ -57,7 +55,16 @@
 #include <getopt.h>
 #endif
 
+#ifdef HAVE_PWD_H
+#include <pwd.h>
+#ifdef HAVE_GRP_H
+#include <grp.h>
+#define RUN_AS_USER
+#endif
+#endif
+
 #include "mrecord.h"
+#include "mhistory.h"
 #include "mstate.h"
 #include "mlocale.h"
 #include "mconfig.h"
@@ -65,242 +72,16 @@
 #include "mdatatypes.h"
 #include "misc.h"
 #include "datatypes/webhist/datatype.h"
+#include "datatypes/mailhist/datatype.h"
 #include "datatypes/state/datatype.h"
+#include "datatypes/query/datatype.h"
+#include "datatypes/string/datatype.h"
+#include "datatypes/count/datatype.h"
 
-#define HISTORY_FILE_VERSION	"0.2"
+#include "mresolver.h"
 
 #define MSTATE_WRITE 1
 
-int history_write(mconfig *conf, mlist *l, char *subpath) {
-	char filename[255];
-	gzFile *fd;
-	
-	sprintf(filename, "%s%s%s/mla.history.xml", 
-		conf->outputdir ? conf->outputdir : ".", 
-		subpath ? "/" : "",
-		subpath ? subpath : "");
-	
-	if ((fd = gzopen(filename, "wb")) == NULL) {
-		fprintf(stderr, "%s.%d: can't open %s: %s\n", __FILE__, __LINE__, filename, strerror(errno));
-		return -1;
-	}
-	
-	gzprintf(fd, "<?xml version=\"1.0\"?>\n");
-	gzprintf(fd, "<!DOCTYPE history PUBLIC \"-//MLA//history\" \"mla.history.dtd\">\n");
-	gzprintf(fd, "<history version=\"%s\" package=\"%s\">\n",
-		HISTORY_FILE_VERSION, PACKAGE);
-	
-	mlist_write(fd, l);
-	
-	gzprintf(fd, "</history>\n");
-	gzclose(fd);
-	
-	return 0;
-}
-
-static xmlSAXHandler mlaSAXHandler = {
-	NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL,
-		NULL
-};
-
-void history_startElement(void *user_data, const xmlChar *name, const xmlChar **attrs) {
-	int i;
-	mstate_stack *m = user_data;
-	
-	/* level 0 <state> */
-	m->st_depth = 0;
-	
-	/* setup the array if we are here the first time */
-	if (m->st[0].id == -1) {
-		for (i = 0; i < M_STATE_ST_ELEM; i++) {
-			m->st[i].id = 0;
-			m->st[i].data = NULL;
-		}
-	}
-#if 0
-	fprintf(stderr, "--> %s\n", name);
-#endif 
-#if 0
-	fprintf(stderr, "o->");
-	for (i = 0; i < m->st_depth_max + 2; i++) {
-		fprintf(stderr, " %d", m->st[i].id);
-	}
-	fprintf(stderr, "\n");
-#endif	
-	/* the first element */
-	if (m->st[0].id == 0) {
-		/* should be state */
-		if (0 == strcmp(name, "history")) {
-			m->st[0].id = 1;
-			/*m->st_depth_max++;*/
-		} else {
-			exit(-1);
-		}
-	} else {
-		/* level 1 <web>, <global> */
-		m->st_depth = 1;
-		if (m->st[m->st_depth].id == 0) {
-			m->st_depth_max++;
-			m->st[m->st_depth].id = 1;
-		}
-		if (mdata_insert_value(m, M_TAG_BEGIN, &(m->state), M_DATA_FIELDTYPE_LIST, name, attrs)) {
-			M_DEBUG0(M_DEBUG_LEVEL_ERRORS, M_DEBUG_SECTION_INIT, M_DEBUG_LEVEL_ERRORS,
-				 "read failed - going down hard\n");
-			exit(-1);
-		}
-	}
-#if 0
-	fprintf(stderr, "o-<");
-	for (i = 0; i < m->st_depth_max + 2; i++) {
-		fprintf(stderr, " %d", m->st[i].id);
-	}
-	fprintf(stderr, "\n");
-#endif
-}
-
-void history_endElement(void *user_data, const xmlChar *name) {
-	int i;
-	mstate_stack *m = user_data;
-	
-	/* level 0 <state> */
-	m->st_depth = 0;
-	
-	/* setup the array if we are here the first time */
-	if (m->st[0].id == -1) {
-		for (i = 0; i < M_STATE_ST_ELEM; i++) {
-			m->st[i].id = 0;
-			m->st[i].data = NULL;
-		}
-	}
-	
-#if 0
-	fprintf(stderr, "<-- %s\n", name);
-#endif
-	
-	/* the first element */
-	if (m->st[0].id != 1) {
-		/* should be history */
-		exit(-1);
-	} else {
-		/* level 1 <web>, <global> */
-		m->st_depth++;
-		if (m->st[m->st_depth].id != 0) {
-			m->st_depth++;
-			if (mdata_insert_value(m, M_TAG_END, &(m->state), M_DATA_FIELDTYPE_LIST, name, NULL)) {
-				M_DEBUG0(M_DEBUG_LEVEL_ERRORS, M_DEBUG_SECTION_INIT, M_DEBUG_LEVEL_ERRORS,
-					 "read failed - going down hard\n");
-				exit(-1);
-			}
-			if (0) {
-				m->st[m->st_depth].id = 0;
-				m->st_depth_max--;
-			}
-		}
-	}
-}
-
-void history_characters(void *user_data, const xmlChar *name, int len) {
-	int i;
-	mstate_stack *m = user_data;
-	
-	xmlChar *s;
-	
-	/* level 0 <state> */
-	m->st_depth = 0;
-	
-	/* setup the array if we are here the first time */
-	if (m->st[0].id == -1) {
-		for (i = 0; i < M_STATE_ST_ELEM; i++) {
-			m->st[i].id = 0;
-			m->st[i].data = NULL;
-		}
-	}
-	
-#if 0	
-	fprintf(stderr, "--| %s\n", s);
-#endif	
-	/* the first element */
-	if (m->st[0].id != 1) {
-		/* should be history */
-		exit(-1);
-	} else {
-		s = malloc((len + 1) * sizeof(xmlChar));
-		strncpy(s, name, len);
-		s[len] = '\0';
-		/* level 1 <web>, <global> */
-		m->st_depth++;
-		if (m->st[m->st_depth].id != 0) {
-			m->st_depth++;
-			if (mdata_insert_value(m, M_TAG_TEXT, &(m->state), M_DATA_FIELDTYPE_LIST, s, NULL)) {
-				M_DEBUG0(M_DEBUG_LEVEL_ERRORS, M_DEBUG_SECTION_INIT, M_DEBUG_LEVEL_ERRORS,
-					 "read failed - going down hard\n");
-				exit(-1);
-			}
-		}
-	}
-	
-	free(s);
-}
-
-
-int history_read(mconfig *conf, mlist *l, char *subpath) {
-	char filename[255];
-	mstate_stack m;
-	struct stat s;
-
-	sprintf(filename, "%s%s%s/mla.history.xml", 
-		conf->outputdir ? conf->outputdir : ".", 
-		subpath ? "/" : "",
-		subpath ? subpath : "");
-	
-	mlaSAXHandler.startElement = history_startElement;
-	mlaSAXHandler.endElement = history_endElement;
-	mlaSAXHandler.characters = history_characters;
-	mlaSAXHandler.warning = xmlParserWarning;
-	mlaSAXHandler.error = xmlParserError;
-	mlaSAXHandler.fatalError = xmlParserError;
-	
-	m.ext_conf = conf;
-	m.st[0].id = -1;
-	m.state = l;
-	m.st_depth_max = 0;
-	m.st_depth = 0;
-	
-	if (0 == stat(filename, &s)) {
-		xmlSAXUserParseFile(&mlaSAXHandler, &m, filename);
-	} else {
-		M_DEBUG1(M_DEBUG_LEVEL_ERRORS, M_DEBUG_SECTION_INIT, M_DEBUG_LEVEL_ERRORS,
-			 "stating history-file '%s' failed\n", filename);
-	}
-	
-	return 0;
-}
-
 void show_version() {
 	printf("%s %s\n", PACKAGE, VERSION);
 	fflush(stdout);
@@ -313,16 +94,19 @@ void show_header() {
 
 void show_usage() {
 	show_header();
-	
 
-#ifdef HAVE_GETOPT_LONG
 	printf("Options:\n");
 	printf(" -c <configfile>    filename of the configfile\n");
+	printf(" -r                 force execution as root\n");
+	printf(" -o <section>:<key>=<value>\n");
+	printf("                    set a config-option in section <section>\n");
+#ifdef RUN_AS_USER
+	printf(" -u <username>      force execution as username (need -r)\n");
+#endif
+#ifdef HAVE_GETOPT_LONG
 	printf(" -h --help          this help screen\n");
 	printf(" -v --version       display version\n");
 #else
-	printf("Options:\n");
-	printf(" -c <configfile>    filename of the configfile\n");
 	printf(" -h                 this help screen\n");
 	printf(" -v                 display version\n");
 #endif
@@ -332,29 +116,35 @@ int restore_internal_state (mconfig *conf, mlist *state_list) {
 	char *fn;
 	char buf[255];
 	FILE *split_file;
+	int empty = 1;
 	
-	if (conf->outputdir == NULL) {
-		return -1;
-	}
-	
-	fn = malloc(strlen(conf->outputdir) + strlen("/modlogan.statefiles") + 1);
-	strcpy(fn, conf->outputdir);
-	strcat(fn, "/modlogan.statefiles");
-	
-	if ((split_file = fopen(fn, "r")) == NULL) {
-		free(fn);
+	if (conf->statedir == NULL) {
+		fprintf(stderr, "%s.%d: global:statdir is empty\n", __FILE__, __LINE__);
 		return -1;
 	}
 
+	fn = malloc(strlen(conf->statedir) + strlen("/modlogan.statefiles") + 1);
+	strcpy(fn, conf->statedir);
+	strcat(fn, "/modlogan.statefiles");
+
+	if ((split_file = fopen(fn, "r")) == NULL) {
+		fprintf(stderr, "%s.%d: can't open %s: %s, first run ?\n", __FILE__, __LINE__, fn, strerror(errno));
+		
+		free(fn);
+		return 0;
+	}
+	
 	while (fgets(buf, sizeof(buf), split_file)) {
 		mstate *state = NULL;
 		mlist *hist = mlist_init();
 		
+		empty = 0;
+		
 		/* remove the \n */
 		buf[strlen(buf)-1] = '\0';
-		
+
 		M_DEBUG2(conf->debug_level, M_DEBUG_SECTION_INIT, M_DEBUG_LEVEL_VERBOSE,
-			"reading statefile from %s/%s/mla.state.xml\n", conf->outputdir, buf);
+			"reading statefile from %s/%s/mla.state.xml\n", conf->statedir, buf);
 
 		if (history_read(conf, hist, buf)) {
 			fprintf(stderr, "%s.%d: hist failed ?\n", __FILE__, __LINE__);
@@ -364,37 +154,49 @@ int restore_internal_state (mconfig *conf, mlist *state_list) {
 
 		if (conf->incremental) {
 			state = mstate_init();
-			
+
 			if (mstate_read(conf, state, 0, 0, buf)) {
 				mstate_free(state);
 				state = NULL;
 			}
 		}
-				
+
 		if (!hist || (conf->incremental && !state)) {
 			if (hist)	mlist_free(hist);
 			else fprintf(stderr, "%s.%d: reading history failed\n", __FILE__, __LINE__);
-			
+
 			if (state)	mstate_free(state);
 			else if (conf->incremental) fprintf(stderr, "%s\n", _("Reading State file failed"));
-			
+
 			state = NULL;
 			hist = NULL;
 		} else {
 			mdata *data = mdata_State_create(buf, state, hist);
 			mlist_insert(state_list, data);
-					
+
 			M_DEBUG1(conf->debug_level, M_DEBUG_SECTION_INIT, M_DEBUG_LEVEL_VERBOSE,
 				"state restored for '%s'\n", buf);
-		} 
+		}
 	}
-
+	
 	fclose(split_file);
+	
+	if (empty) {
+		fprintf(stderr, "%s.%d: \n"
+			"  modlogan.statefiles (%s) was found, but the file is empty\n"
+			"  looks like modlogan died in the previous run while writing the output\n", 
+			__FILE__, __LINE__, fn);
+		free(fn);
+		return -1;
+	} 
+	
 	free(fn);
 	
 	return 0;
 }
 
+
+
 typedef struct {
 	mlogrec *record;
 	int rec_state;
@@ -403,42 +205,129 @@ typedef struct {
 
 int get_next_record (mconfig *conf, mlogrec *rec) {
 	mplugin *func;
-	int i;
+	int i, j;
 	int state = M_RECORD_EOF;
+	int last_rec_state = M_RECORD_NO_ERROR;
 	static mgetrec **recs = NULL;
+	static mgetrec **readahead = NULL;
 	time_t t = -1;
 	int ndx = -1;
-	
-	if (recs == NULL) {
-		/* a small mem leak. fix it if you like */
+	time_t rt = -1;
+	int rndx = -1;
+
+	if (recs == NULL && conf->plugin_count) {
 		recs = malloc(sizeof(mgetrec *) * conf->plugin_count);
 		for (i = 0; i < conf->plugin_count; i++) {
-			/* a leak - but we don't care about this one either */
 			recs[i] = malloc(sizeof(mgetrec));
 			recs[i]->record = mrecord_init();
 			recs[i]->rec_state = M_RECORD_EOF;
 		}
+
 	}
-	
-	for (i = 0; i < conf->plugin_count; i++) {
-		func = ((mplugin **)(conf->plugins))[i];
-		conf->plugin_conf = func->config;
-		
-		if (func->get_next_record) {
-			if (recs[i]->record->ext_type == M_RECORD_TYPE_UNSET) {
-				recs[i]->rec_state = func->get_next_record(conf, recs[i]->record);
+
+	if (readahead == NULL) {
+		readahead = malloc(sizeof(mgetrec *) * conf->read_ahead_limit);
+		for (i = 0; i < conf->read_ahead_limit; i++) {
+			readahead[i] = malloc(sizeof(mgetrec));
+			readahead[i]->record = mrecord_init();
+			readahead[i]->rec_state = M_RECORD_EOF;
+		}
+
+	}
+
+	for (j = 0; j < conf->read_ahead_limit; j++) {
+		if (last_rec_state != M_RECORD_EOF &&
+		    readahead[j]->record->timestamp == 0) {
+			/* is free */
+
+			ndx = -1;
+			t = -1;
+			/* check the plugins for new records */
+			for (i = 0; i < conf->plugin_count; i++) {
+				func = ((mplugin **)(conf->plugins))[i];
+				conf->plugin_conf = func->config;
+
+				if (func->get_next_record) {
+					if (recs[i]->record->timestamp == 0) {
+						recs[i]->rec_state = func->get_next_record(conf, recs[i]->record);
+
+						last_rec_state = recs[i]->rec_state;
+
+						switch(recs[i]->rec_state) {
+						case M_RECORD_NO_ERROR:
+							/* start the resolver for this record */
+							resolver_start(conf, recs[i]->record);
+							break;
+						case M_RECORD_SKIPPED:
+						case M_RECORD_CORRUPT:
+						case M_RECORD_IGNORED:
+						case M_RECORD_HARD_ERROR:
+							return recs[i]->rec_state;
+						case M_RECORD_EOF:
+							break;
+						}
+					}
+
+					if (t == -1 || recs[i]->record->timestamp < t) {
+						t = recs[i]->record->timestamp;
+						ndx = i;
+					}
+				}
 			}
-		
-			if (t == -1 || recs[i]->record->timestamp < t) {
-				t = recs[i]->record->timestamp;
-				ndx = i;
+
+			/* got a new record */
+			if (ndx != -1) {
+				mrecord_move(readahead[j]->record, recs[ndx]->record);
+				readahead[j]->rec_state = recs[ndx]->rec_state;
+#if 0
+				fprintf(stderr, "%s.%d: readahead: filled slot %d, %d\n",
+					__FILE__, __LINE__,
+					j, readahead[j]->record->ext_type);
+#endif
+			} else {
+#if 0
+				fprintf(stderr, "%s.%d: no record for me\n", __FILE__, __LINE__);
+#endif
 			}
 		}
+
+		/* select the oldest record from our readahead storage */
+		if (readahead[j]->record->ext_type != M_RECORD_TYPE_UNSET) {
+			if  (rt == -1 || readahead[j]->record->timestamp < rt) {
+				rt = readahead[j]->record->timestamp;
+				rndx = j;
+			} else {
+#if 0
+				fprintf(stderr, "%s.%d: %d - %d\n", __FILE__, __LINE__, rt, readahead[j]->record->timestamp);
+#endif
+			}
+		} else {
+#if 0
+			fprintf(stderr, "%s.%d - no record\n", __FILE__, __LINE__);
+#endif
+		}
 	}
-	
-	if (ndx != -1) {
-		mrecord_move(rec, recs[ndx]->record);
-		state = recs[ndx]->rec_state;
+
+	if (rndx != -1) {
+		mrecord_move(rec, readahead[rndx]->record);
+		state = readahead[rndx]->rec_state;
+	} else {
+		/* got no record, kill myself */
+		for (i = 0; i < conf->plugin_count; i++) {
+			mrecord_free(recs[i]->record);
+			free(recs[i]);
+		}
+
+		free(recs);
+		recs = NULL;
+
+		for (i = 0; i < conf->read_ahead_limit; i++) {
+			mrecord_free(readahead[i]->record);
+			free(readahead[i]);
+		}
+
+		free(readahead);
+		readahead = NULL;
 	}
 
 	return state;
@@ -447,11 +336,25 @@ int get_next_record (mconfig *conf, mlogrec *rec) {
 int generate_monthly_output(mconfig *conf, mstate *state, const char *subpath) {
 	int i;
 	mplugin *func;
+
+	if (subpath) {
+		/* set the splitter */
+		mdata *data = mdata_String_create("splitby", subpath);
+		
+		if (data) {
+			mhash_insert_replace(conf->variables, data);
+		}
+	}
+
+	/* call all output-generators */
 	for (i = 0; i < conf->plugin_count; i++) {
 		func = ((mplugin **)(conf->plugins))[i];
 		conf->plugin_conf = func->config;
-		
-		if (func->gen_report &&	func->gen_report(conf, state, subpath)) return -1;
+
+		if (func->gen_report) {
+			if (func->gen_report(conf, state, subpath))
+				return -1;
+		}
 	}
 
 	return 0;
@@ -459,6 +362,15 @@ int generate_monthly_output(mconfig *conf, mstate *state, const char *subpath) {
 int generate_history_output(mconfig *conf, mlist *history, const char *subpath) {
 	int i;
 	mplugin *func;
+
+	if (subpath) {
+		mdata *data = mdata_String_create("splitby", subpath);
+		
+		if (data) {
+			mhash_insert_replace(conf->variables, data);
+		}
+	}
+
 	for (i = 0; i < conf->plugin_count; i++) {
 		func = ((mplugin **)(conf->plugins))[i];
 		conf->plugin_conf = func->config;
@@ -470,27 +382,28 @@ int generate_history_output(mconfig *conf, mlist *history, const char *subpath)
 }
 
 void dump (mlist *state_list) {
-	mlist *s = state_list;
+	mlist *s;
 	int i;
-	
-	while (s) {
+
+	for (s = state_list; s; s = s->next) {
 		mstate_web *w;
 		if (!s->data) return;
-		
+
 		w = ((mstate *)(s->data->data.state.state))->ext;
-	
+		
 		for (i = 0; i < 31; i++) {
 			fprintf(stderr, "dump_list (%p)->'%s': %d: %ld\n", s, s->data->key, i, w->days[i].hits);
 		}
-		
-		s = s->next;
 	}
 }
 
 int insert_record(mconfig *conf, mlist *state_list, mlogrec *record) {
 	int i = 0;
 	mplugin *func;
-	
+
+
+	resolver_finish(conf, record);
+
 	for (i = 0; i < conf->plugin_count; i++) {
 		func = ((mplugin **)(conf->plugins))[i];
 		conf->plugin_conf = func->config;
@@ -503,12 +416,74 @@ int insert_record(mconfig *conf, mlist *state_list, mlogrec *record) {
 	return 0;
 }
 
+#ifdef RUN_AS_USER
+/* Change to user privileges. May be only called by root.
+ * It returns 0 on success.
+ * -1 on any error.
+ */
+static int change_to_user(char *user) {
+	struct passwd *pwd;
+
+	/* Get user info */
+	pwd = getpwnam (user);
+	if (!pwd) return -1;
+
+	/* We don't want to change to root user, -r is here for that. */
+	if (!pwd->pw_uid) return -1;
+
+#ifdef HAVE_SETGROUPS
+#ifdef HAVE_GETGROUPLIST
+#ifndef NGROUPS
+#define NGROUPS 32
+#endif
+	{
+		gid_t *groups;
+		int ngroups = NGROUPS;
+
+		/* Get user group list */
+		groups = (gid_t *) malloc (ngroups * sizeof (gid_t));
+		if (!groups) return -1;
+		if (getgrouplist(user, pwd->pw_gid, groups, &ngroups) == -1) {
+			groups = (gid_t *) realloc (groups, ngroups * sizeof (gid_t));
+			if (!groups) return -1;
+			if (getgrouplist(user, pwd->pw_gid, groups, &ngroups) == -1) {
+				free(groups);
+				return -1;
+			}
+		}
+
+		/* Set additionnal groups */
+		if (setgroups(ngroups, groups) == -1) {
+			free(groups);
+			return -1;
+		}
+		free(groups);
+
+	}
+#undef NGROUPS
+#else
+	/* Remove any additionnal group */
+	if (setgroups(0, NULL) == -1) return -1;
+#endif /* HAVE_GETGROUPLIST */
+#endif /* HAVE_SETGROUPS */
+
+	/* Set main group */
+	if (setgid (pwd->pw_gid) == -1) return -1;
+	/* Set user */
+	if (setuid (pwd->pw_uid) == -1) return -1;
+
+	/* Re-check ... */
+	if (getuid() != pwd->pw_uid || geteuid() != pwd->pw_uid) return -1;
+	return 0;
+}
+#endif /* RUN_AS_USER */
+
 int main (int argc, char **argv) {
 	mconfig *conf = NULL;
 	mlogrec *rec = NULL;
 	mlist *state_list = NULL;
 	mlist *l;
-	
+
 	int ret, l_month = -1, l_year = -1, l_hour = -1, l_day = -1, i;
 	char *conf_filename = NULL;
 	long lines = 0, lines_corrupt = 0, lines_skipped = 0, lines_ignored = 0;
@@ -516,49 +491,62 @@ int main (int argc, char **argv) {
 	int root_check = 1;
 
 	struct tm *tm;
-	
+
 	time_t l_stamp = -1;
 	mtimer process_timer, post_process_timer, parse_timer, setup_timer;
-	
-#ifdef HAVE_GETOPT_LONG	
+
+#ifdef RUN_AS_USER
+	char *run_as_user = NULL;
+#define CMD_OPTIONS "c:hvro:u:"
+#else
+#define CMD_OPTIONS "c:hvro:"
+#endif
+
+#ifdef HAVE_GETOPT_LONG
 	int option_index = 0;
 	static struct option long_options[] = {
-		{"help", 0, NULL, 0},
-		{"version", 0, NULL, 0},
-		{"root", 0, NULL, 0},
-		{NULL, 0, NULL, 0}
+		{ "help", 0, NULL, 'h' },
+		{ "version", 0, NULL, 'v' },
+		{ "root", 0, NULL, 'r' },
+		{ NULL, 0, NULL, 0 }
 	};
 #endif
-	
+
 	MTIMER_RESET(process_timer);
 	MTIMER_RESET(post_process_timer);
 	MTIMER_RESET(parse_timer);
 	MTIMER_RESET(setup_timer);
 	
-#ifdef HAVE_GETOPT_LONG	
-	while ((i = getopt_long(argc,argv,"c:hvr", 
-				long_options, &option_index)) != -1) {
-		
-		/* map the long options to the short ones */
-		if (i == 0) {
-			switch(option_index) {
-			case 0: i = 'h'; break;
-			case 1: i = 'v'; break;
-			case 2: i = 'r'; break;
-			default:
-				show_usage();
-				exit( EXIT_FAILURE);
-			}
-		}
-#if 0
-		/* for jed and it formating problems */
+	if (!(conf = mconfig_init())) {
+		fprintf(stderr, "%s.%d: init config failed\n", __FILE__, __LINE__);
+		return -1;
 	}
-#endif
+	
+	/* 
+	 * command line options
+	 * 
+	 * -h/--help    help
+	 * -r/--root    allow root
+	 * -u <uid>     set new uid
+	 * -v/--version version
+	 * -c <file>    configfile
+	 * 
+	 * -o <section>:<key>=<value>
+	 */
+	
+	opterr = 0;
+
+	while (
+#ifdef HAVE_GETOPT_LONG
+	       (i = getopt_long(argc, argv, CMD_OPTIONS,
+				long_options, &option_index)) != -1
+	       
 #else
-	while ((i = getopt(argc,argv,"c:hvr")) != EOF) { 
-#endif		
+	       (i = getopt(argc,argv,CMD_OPTIONS)) != EOF
+#endif
+	) {
 		switch(i) {
-		case 'c': 
+		case 'c':
 			conf_filename = optarg;
 			break;
 		case 'h':
@@ -570,16 +558,29 @@ int main (int argc, char **argv) {
 		case 'r':
 			root_check = 0;
 			break;
+#ifdef RUN_AS_USER
+		case 'u':
+			run_as_user = optarg;
+			break;
+#endif
+		case 'o': {
+			mdata *data = mdata_Count_create(optarg, 1, 0);
+			/* add option to the config-file handling */
+			
+			mlist_insert_sorted(conf->cmdlineoptions, data);
+			
+			break;
+		}
 		default:
 			show_usage();
 			exit( EXIT_FAILURE);
 		}
 	}
-		
+
 	show_header();
-	
-	if (root_check) {
-		if (getuid() == 0 || geteuid() == 0) {
+
+	if (getuid() == 0 || geteuid() == 0) {
+		if (root_check) {
 			fprintf(stderr, "ModLogAn detected that it is running with root permissions.\n"
 				"As root permissions are not directly required to run modlogan, \n"
 				"modlogan decided do stop working with root permission.\n\n"
@@ -591,49 +592,63 @@ int main (int argc, char **argv) {
 				"use the -r/--root switch which will disable this test\n");
 			exit(-1);
 		}
+#ifdef RUN_AS_USER
+		if (run_as_user) {
+			if (change_to_user(run_as_user)) {
+				fprintf(stderr, "ModLogAn failed to change to user '%s'. Exiting.\n", run_as_user);
+				exit(-1);
+			} else {
+				fprintf(stderr, "ModLogAn is running as user '%s'.\n", run_as_user);
+			}
+		} else
+#endif
+			fprintf(stderr, "WARNING: ModLogAn is running as user 'root'.\n");
+
 	}
-	
+
 	init_locale(conf);
-	
+
 	MTIMER_START(setup_timer);
-	
-	if (!(conf = mconfig_init(conf_filename))) {
+
+	if (mconfig_read(conf, conf_filename)) {
 		fprintf(stderr, "%s.%d: reading configfile failed - going down hard\n", __FILE__, __LINE__);
 		return -1;
 	}
-	
+
 	if (conf->show_options) {
 		/* we only have to show some options */
-		
+
 		mconfig_free(conf);
-		
+
 		return 0;
 	}
-	
+
 	state_list = mlist_init();
 	rec = mrecord_init();
 
-	restore_internal_state(conf, state_list);
-	
+	if (restore_internal_state(conf, state_list)) {
+		return -1;
+	}
+
 	if (conf->incremental) {
 		l = state_list;
-		
+
 		while (l) {
 			mdata *data = l->data;
 			mstate *int_state;
 			if (!data) break;
-			
+
 			int_state = data->data.state.state;
 			if (int_state->timestamp > l_stamp) {
 				l_stamp = int_state->timestamp;
 				l_year = int_state->year-1900;
 				l_month = int_state->month-1;
 			}
-			
+
 			l = l->next;
 		}
 	}
-		
+
 	if (l_stamp != -1) {
 		M_DEBUG1(conf->debug_level, M_DEBUG_SECTION_INIT, M_DEBUG_LEVEL_VERBOSE,
 			 "Ignoring all records before %s", ctime(&l_stamp));
@@ -641,29 +656,30 @@ int main (int argc, char **argv) {
 
 	M_DEBUG0(conf->debug_level, M_DEBUG_SECTION_INIT, M_DEBUG_LEVEL_NONE,
 		"startup - finished\n");
-	
+
 	MTIMER_STOP(setup_timer);
 	MTIMER_CALC(setup_timer);
-	
+
 	printf("[");
-	
+
+	/* mainloop */
 	MTIMER_START(parse_timer);
 	while ((ret = get_next_record(conf, rec)) != M_RECORD_EOF) {
 		MTIMER_STOP(parse_timer);
 		MTIMER_CALC(parse_timer);
 
 		MTIMER_START(process_timer);
-		
+
 		lines++;
 		if (ret == M_RECORD_NO_ERROR) {
 			/* HACK */
 			if (rec->timestamp == 0) {
-				fprintf(stderr, "%s.%d: line %ld returned no timestamp !! something strange will happen. Going down hard\n", 
+				fprintf(stderr, "%s.%d: line %ld returned no timestamp !! something strange will happen. Going down hard\n",
 					__FILE__, __LINE__,
 					lines);
 				return -1;
-			} 
-			
+			}
+
 			/* set the first timestamp */
 			if (l_month == -1) {
 				tm = localtime(&(rec->timestamp));
@@ -673,7 +689,13 @@ int main (int argc, char **argv) {
 				l_hour = tm->tm_hour;
 			}
 
-			if (l_stamp > rec->timestamp) {
+			/* is the current timestamp below the timestamp
+			 * of the last successfully parsed record ?
+			 */
+			if (rec->timestamp < l_stamp) {
+				/* skip the record if we are in incremental mode 1 and no
+				 * record has been parsed successfully yet
+				 */
 				if (first_valid_line == 0 && conf->incremental == 1) {
 					lines_skipped++;
 					MTIMER_STOP(process_timer);
@@ -683,42 +705,62 @@ int main (int argc, char **argv) {
 				}
 				M_DEBUG2(conf->debug_level, M_DEBUG_SECTION_MAINLOOP, M_DEBUG_LEVEL_VERBOSE,
 					"Out of Sequence - ignoring timestamp: %ld > %ld\n", l_stamp, rec->timestamp);
-				
+
 				rec->timestamp = l_stamp;
 			}
-			
+
 			tm = localtime(&(rec->timestamp));
-			
+
 			/* set last day of month */
 			l_day = tm->tm_mday;
 
 			/* generate report if we have reached a threshold (semi-online mode) */
-			if (conf->gen_report_threshold > 0 && 
+			if (conf->gen_report_threshold > 0 &&
 				(lines - lines_corrupt + 1) % conf->gen_report_threshold == 0) {
-				
+
 				MTIMER_STOP(process_timer);
 				MTIMER_CALC(process_timer);
-				
+
 				printf("]\n");
-				
+
 				MTIMER_START(post_process_timer);
 				M_DEBUG2(conf->debug_level, M_DEBUG_SECTION_MAINLOOP, M_DEBUG_LEVEL_WARNINGS,
 					"o(t) writing month %02i - %04i\n", l_month+1, l_year+1900);
-				
-				l = state_list;
-				
-				while (l) {
+
+
+				for (l = state_list; l; l = l->next) {
 					mdata *data = l->data;
 					mdata *histdata = NULL;
-					
+					mstate *state;
+
 					if (!data) break;
-					
-					histdata = mdata_WebHist_create_by_state(data->data.state.state);
-				
+
+					state = data->data.state.state;
+
+					switch (state->ext_type) {
+					case M_STATE_TYPE_WEB:
+						histdata = mdata_WebHist_create_by_state(state);
+						break;
+					case M_STATE_TYPE_MAIL:
+						histdata = mdata_Mailhist_create_by_state(state);
+						break;
+					case M_STATE_TYPE_UNSET:
+						continue;
+					default:
+						M_DEBUG1(conf->debug_level, M_DEBUG_SECTION_MAINLOOP, M_DEBUG_LEVEL_ERRORS,
+							 "Unknown state-type: %d\n", state->ext_type);
+						return (-1);
+					}
+
 					mlist_insert_replace(data->data.state.history, histdata);
+
+					/* sort the history list */
+					data->data.state.history = mlist_sort_full_by_string(data->data.state.history);
+					
 #if MSTATE_WRITE
-					if (0 != mstate_write(conf, data->data.state.state, 
-							      M_STATE_WRITE_BY_MONTH, 
+					
+					if (0 != mstate_write(conf, data->data.state.state,
+							      M_STATE_WRITE_BY_MONTH,
 							      *(data->key) ? data->key : NULL)) {
 						fprintf(stderr, "%s.%d: writing state file failed\n", __FILE__, __LINE__);
 						return -1;
@@ -733,47 +775,74 @@ int main (int argc, char **argv) {
 						fprintf(stderr, "%s.%d: history generation failed - going down hard\n", __FILE__, __LINE__);
 						return -1;
 					}
-	
-					l = l->next;
+
 				}
 				MTIMER_STOP(post_process_timer);
 				MTIMER_CALC(post_process_timer);
-				
-				
+
+
 				printf("[");
 				MTIMER_START(process_timer);
 			}
-			
+
 			/* we have left a month, generate the reports */
 			if (tm->tm_mon != l_month || tm->tm_year != l_year) {
-				
+
 				MTIMER_STOP(process_timer);
 				MTIMER_CALC(process_timer);
-				
+
 				printf("]\n");
-				
+
 				MTIMER_START(post_process_timer);
-				
+
 				M_DEBUG2(conf->debug_level, M_DEBUG_SECTION_MAINLOOP, M_DEBUG_LEVEL_WARNINGS,
 					"o writing month %02i - %04i\n", l_month+1, l_year+1900);
-				
-				
-				l = state_list;
-				
-				while (l) {
+
+
+				for (l = state_list; l; l = l->next) {
 					mdata *data = l->data;
 					mdata *histdata = NULL;
-					
+					mstate *state;
+
 					if (!data) break;
-					
-					histdata = mdata_WebHist_create_by_state(data->data.state.state);
-				
+
+					state  = data->data.state.state;
+
+					switch (state->ext_type) {
+					case M_STATE_TYPE_WEB:
+						histdata = mdata_WebHist_create_by_state(state);
+						break;
+					case M_STATE_TYPE_MAIL:
+						histdata = mdata_Mailhist_create_by_state(state);
+						break;
+					case M_STATE_TYPE_UNSET:
+						if (state->ext) {
+							M_DEBUG0(conf->debug_level, M_DEBUG_SECTION_MAINLOOP, M_DEBUG_LEVEL_ERRORS,
+								 "no ext_type, but ext is set\n");
+							return -1;
+						}
+						continue;
+					default:
+						M_DEBUG1(conf->debug_level, M_DEBUG_SECTION_MAINLOOP, M_DEBUG_LEVEL_ERRORS,
+							 "Unknown state-type: %d\n", state->ext_type);
+						return (-1);
+					}
+
 					mlist_insert_replace(data->data.state.history, histdata);
+
+					/* sort the history list */
+					data->data.state.history = mlist_sort_full_by_string(data->data.state.history);
 					
 					if (conf->debug_level > 3)
 						fprintf(stderr, "o(t) -- state written: (...)/%s/\n", *(data->key) ? data->key : ".");
 #if MSTATE_WRITE
-					mstate_write(conf, data->data.state.state, M_STATE_WRITE_BY_MONTH, *(data->key) ? data->key : NULL);
+					if (0 != mstate_write(conf, data->data.state.state,
+							      M_STATE_WRITE_BY_MONTH,
+							      *(data->key) ? data->key : NULL)) {
+						fprintf(stderr, "%s.%d: writing state file failed\n", __FILE__, __LINE__);
+						return -1;
+					}
+					history_write(conf, data->data.state.history, *(data->key) ? data->key : NULL);
 #endif
 					if (0 != generate_monthly_output(conf, data->data.state.state, *(data->key) ? data->key : NULL)) {
 						fprintf(stderr, "%s.%d: output generation failed - no data this month?\n", __FILE__, __LINE__);
@@ -783,24 +852,22 @@ int main (int argc, char **argv) {
 						fprintf(stderr, "%s.%d: history generation failed - going down hard\n", __FILE__, __LINE__);
 						return -1;
 					}
-					
+
 					mstate_free(data->data.state.state);
 					data->data.state.state = mstate_init();
-					
-					l = l->next;
 				}
-				
+
 				tm = localtime(&(rec->timestamp));
-			
+
 				l_month = tm->tm_mon;
 				l_year = tm->tm_year;
-				
+
 				MTIMER_STOP(post_process_timer);
 				MTIMER_CALC(post_process_timer);
-				
+
 				for (i = 0; i < ((lines % 50000) / 1000); i++) printf(" ");
 				printf("[");
-				
+
 				MTIMER_START(process_timer);
 			}
 
@@ -808,9 +875,9 @@ int main (int argc, char **argv) {
 				fprintf(stderr, "%s.%d: inserting record failed - going down hard\n", __FILE__, __LINE__);
 				return -1;
 			}
-			
+
 			first_valid_line = 1;
-			
+
 			l_stamp = rec->timestamp;
 		} else if (ret == M_RECORD_SKIPPED) {
 			lines_skipped++;
@@ -818,36 +885,36 @@ int main (int argc, char **argv) {
 			lines_ignored++;
 		} else if (ret == M_RECORD_HARD_ERROR) {
 			M_DEBUG1(conf->debug_level, M_DEBUG_SECTION_MAINLOOP, M_DEBUG_LEVEL_ERRORS,
-				 " a hard error occured in line %ld - going down hard\n", lines);
-			
+				 "a hard error occured in line %ld - going down hard\n", lines);
+
 			exit(-1);
 		} else {
 			lines_corrupt++;
-			
+
 			M_DEBUG1(conf->debug_level, M_DEBUG_SECTION_MAINLOOP, M_DEBUG_LEVEL_ERRORS,
-				 " parser reported an error in line %ld\n", lines);
+				 "parser reported an error in line %ld\n", lines);
 		}
 		mrecord_reset(rec);
-		
+
 		/* cosmetics */
 		if (lines % 1000 == 0) {
-			printf("."); 
+			printf(".");
 			if (lines % (1000 * 50)== 0) {
 				printf(" %8ld", lines);
 				if (conf->debug_level > 1) {
 					printf(" - %10.2f (%4ld, %4ld)",
-					       (lines)/((MTIMER_GET_USER_MSEC(parse_timer)+MTIMER_GET_USER_MSEC(process_timer))/1000.0), 
+					       (lines)/((MTIMER_GET_USER_MSEC(parse_timer)+MTIMER_GET_USER_MSEC(process_timer))/1000.0),
 					       lines_corrupt,
 					       lines_skipped);
 				}
 				printf("\n ");
 			}
-			fflush(stdout); 
+			fflush(stdout);
 		}
-		
+
 		MTIMER_STOP(process_timer);
 		MTIMER_CALC(process_timer);
-		
+
 		MTIMER_START(parse_timer);
 	}
 
@@ -859,64 +926,122 @@ int main (int argc, char **argv) {
 	MTIMER_START(post_process_timer);
 
 	if (l_month != -1) {
-		int create_bp = 1;
 		FILE *split_file;
 		char *fn;
 		printf(" ");
 
-		if (conf->debug_level > 0) 
+		if (conf->debug_level > 0)
 			printf("writing month %02i - %04i\n",l_month+1, l_year+1900);
-		
+
 		if (conf->debug_level > 1)
 			fprintf(stderr, "%s.%d: writing the last month\n", __FILE__, __LINE__ );
-		
+#if MSTATE_WRITE
 		/* create filename */
-		fn = malloc(strlen(conf->outputdir) + strlen("/modlogan.statefiles") + 1);
-		strcpy(fn, conf->outputdir);
+		fn = malloc(strlen(conf->statedir) + strlen("/modlogan.statefiles") + 1);
+		strcpy(fn, conf->statedir);
 		strcat(fn, "/modlogan.statefiles");
-		split_file = fopen(fn, "w");
-			
+		if (NULL == (split_file = fopen(fn, "w"))) {
+			fprintf(stderr, "%s.%d: %s\n",
+				__FILE__, __LINE__,
+				strerror(errno));
+		} else {
+			fclose(split_file);
+		}
+#endif
+
 		l = state_list;
-				
-		while (l) {
+
+		for (l = state_list; l; l = l->next) {
 			mdata *data = l->data;
 			mdata *histdata = NULL;
-			
+			int faulty = 0;
+			mstate *state;
+
 			if (!data) break;
+
+			state = data->data.state.state;
 			
 			if (conf->debug_level > 1)
-				fprintf(stderr, "%s.%d: subpath = '%s'\n", __FILE__, __LINE__, data->key);
-			
-			histdata = mdata_WebHist_create_by_state(data->data.state.state);
-				
-			mlist_insert_replace(data->data.state.history, histdata);
-#if MSTATE_WRITE
-			mstate_write(conf, data->data.state.state, M_STATE_WRITE_DEFAULT, *(data->key) ? data->key : NULL);
-			history_write(conf, data->data.state.history, *(data->key) ? data->key : NULL);
-#endif
-			if (0 != generate_monthly_output(conf, data->data.state.state, *(data->key) ? data->key : NULL)) {
-				fprintf(stderr, "%s.%d: output generation failed - no data this month?\n", __FILE__, __LINE__);
-				/* return -1; */
+				fprintf(stderr, "%s.%d: subpath = '%s'\n",
+					__FILE__, __LINE__,
+					data->key );
+
+			switch (state->ext_type) {
+			case M_STATE_TYPE_WEB:
+				if (NULL == (histdata = mdata_WebHist_create_by_state(state))) {
+					M_WP();
+					
+					return -1;
+				}
+				break;
+			case M_STATE_TYPE_MAIL:
+				histdata = mdata_Mailhist_create_by_state(state);
+				break;
+			case M_STATE_TYPE_TRAFFIC:
+				break;
+			case M_STATE_TYPE_UNSET:
+				if (state->ext) {
+					M_DEBUG0(conf->debug_level, M_DEBUG_SECTION_MAINLOOP, M_DEBUG_LEVEL_ERRORS,
+						 "no ext_type, but ext is set\n");
+					return -1;
+				}
+				continue;
+			default:
+				M_DEBUG1(conf->debug_level, M_DEBUG_SECTION_MAINLOOP, M_DEBUG_LEVEL_ERRORS,
+					 "Unknown state-type: %d\n", state->ext_type);
+				return (-1);
 			}
-			if (0 != generate_history_output(conf, data->data.state.history, *(data->key) ? data->key : NULL)) {
-				fprintf(stderr, "%s.%d: history generation failed - going down hard\n", __FILE__, __LINE__);
-				return -1;
+#if 0
+			for (hl = data->data.state.history; hl; hl = hl->next) {
+				fprintf(stderr, "%s.%d: %s, %d, %d\n",
+					__FILE__, __LINE__,
+					hl->data->key, hl->data->type,
+					hl->data->data.hist.type);
+			}
+#endif
+			
+			if (histdata) mlist_insert_replace(data->data.state.history, histdata);
+#if 0
+			for (hl = data->data.state.history; hl; hl = hl->next) {
+				fprintf(stderr, "%s.%d: %s, %d, %d\n",
+					__FILE__, __LINE__,
+					hl->data->key, hl->data->type,
+					hl->data->data.hist.type);
+			}
+#endif			
+
+#if MSTATE_WRITE
+			if (!faulty && mstate_write(conf, data->data.state.state, M_STATE_WRITE_DEFAULT, *(data->key) ? data->key : NULL)) {
+				faulty = 1;
+			}
+			if (!faulty && history_write(conf, data->data.state.history, *(data->key) ? data->key : NULL)) {
+				faulty = 1;
+			}
+			if (!faulty) {
+				/* write the current directoy to the split file */
+				if (NULL != (split_file = fopen(fn, "a+"))) {
+					fprintf(split_file, "%s\n", data->key);
+					fclose(split_file);
+				}
+			}
+#endif
+
+			/* generate reports */
+			if (!faulty && 0 != generate_monthly_output(conf, data->data.state.state, *(data->key) ? data->key : NULL)) {
+				fprintf(stderr, "%s.%d: output generation failed - no data this month?\n", __FILE__, __LINE__);
+				faulty = 1;
 			}
 
-			l = l->next;
-			
-			/* disable the bigpicture if we aren't splitting the logs */
-			if (*(data->key) == '\0') create_bp = 0;
-			
-			/* write the current directoy to the split file */
-			fprintf(split_file, "%s\n", data->key);
-			
+			/* generate history */
+			if (!faulty && 0 != generate_history_output(conf, data->data.state.history, *(data->key) ? data->key : NULL)) {
+				fprintf(stderr, "%s.%d: history generation failed - going down hard\n", __FILE__, __LINE__);
+				faulty = 1;
+			}
 		}
-		
-		fclose(split_file);
+
 		free(fn);
 	}
-	
+
 	MTIMER_STOP(post_process_timer);
 	MTIMER_CALC(post_process_timer);
 
@@ -942,15 +1067,15 @@ int main (int argc, char **argv) {
 			MTIMER_GET_SYSTEM_MSEC(post_process_timer)/1000.0);
 
 		printf("%s: %.2f %s (%ld %s, %ld %s, %ld %s, %ld %s)\n",
-		       "Throughput", (lines)/((MTIMER_GET_USER_MSEC(parse_timer)+MTIMER_GET_USER_MSEC(process_timer))/1000.0), 
+		       "Throughput", (lines)/((MTIMER_GET_USER_MSEC(parse_timer)+MTIMER_GET_USER_MSEC(process_timer))/1000.0),
 		       "rec/s",
-		       lines, "records", 
+		       lines, "records",
 		       lines_corrupt, "corrupt records",
 		       lines_skipped, "skipped records",
 		       lines_ignored, "ignored records"
 		       );
 	}
-	
+
 	mconfig_free(conf);
 	return 0;
 }
