@@ -2,7 +2,7 @@
 ** Modular Logfile Analyzer
 ** Copyright 2000 Jan Kneschke <jan@kneschke.de>
 **
-** Homepage: http://www.kneschke.de/projekte/modlogan
+** Homepage: http://www.modlogan.org
 **
 
     This program is free software; you can redistribute it and/or modify
@@ -22,7 +22,7 @@
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 
 **
-** $Id: mlocale.c,v 1.28 2001/10/24 18:19:14 ostborn Exp $
+** $Id: mlocale.c,v 1.36 2003/01/06 10:56:13 ostborn Exp $
 */
 
 #include <stdlib.h>
@@ -37,6 +37,9 @@ int init_locale (mconfig *conf) {
 	setlocale(LC_ALL, "");
 	bindtextdomain (PACKAGE, LOCALEDIR);
 	textdomain (PACKAGE);
+# ifdef HAVE_BIND_TEXTDOMAIN_CODESET
+	bind_textdomain_codeset(PACKAGE,"ISO8859-1");
+# endif
 #endif
 
 	return 0;
@@ -88,7 +91,7 @@ char *mhttpcodes(int status) {
 	}
 }
 
-long long isondx(char *str) {
+long long isondx(const char *str) {
 	int i = 0;
 	int j = 3;
 	if (!str || strlen(str) > 4) return M_RESOLV_UNKNOWN_ERROR;
@@ -99,18 +102,18 @@ long long isondx(char *str) {
 		str++;
 		j--;
 	}
-	
+
 	return i;
 }
 
 #define NDX2(x,y)	x << 24 | y << 16
-#define NDX3(x,y,z)	x << 24 | y << 16 | z << 8 
+#define NDX3(x,y,z)	x << 24 | y << 16 | z << 8
 #define NDX4(x,y,z,a)	x << 24 | y << 16 | z << 8 | a
 
-char *misoname(char *str) {
+char *misoname(const char *str) {
 	switch(isondx(str)) {
 /* .ac */	case NDX2('a','c'):	return _("Ascension Island");
-/* .ad */	case NDX2('a','d'): 	return _("Andorra");	
+/* .ad */	case NDX2('a','d'): 	return _("Andorra");
 /* .ae */	case NDX2('a','e'):	return _("United Arab Emirates");
 /* .af */	case NDX2('a','f'):	return _("Afghanistan");
 /* .ag */	case NDX2('a','g'):	return _("Antigua and Barbuda");
@@ -135,6 +138,7 @@ char *misoname(char *str) {
 /* .bg */	case NDX2('b','g'):	return _("Bulgaria");
 /* .bh */	case NDX2('b','h'):	return _("Bahrain");
 /* .bi */	case NDX2('b','i'):	return _("Burundi");
+/* .biz */	case NDX3('b','i','z'):	return _("Business (.biz)");
 /* .bj */	case NDX2('b','j'):	return _("Benin");
 /* .bm */	case NDX2('b','m'):	return _("Bermuda");
 /* .bn */	case NDX2('b','n'):	return _("Brunei Darussalam");
@@ -159,6 +163,7 @@ char *misoname(char *str) {
 /* .cn */	case NDX2('c','n'):	return _("China");
 /* .co */	case NDX2('c','o'):	return _("Colombia");
 /* .com */	case NDX3('c','o','m'):	return _("US Commercial");
+/* .coop */	case NDX4('c','o','o','p'):	return _("Cooperation (.coop)");
 /* .cr */	case NDX2('c','r'):	return _("Costa Rica");
 /* .cs */	case NDX2('c','s'):	return _("Czechoslovakia (former)");
 /* .cu */	case NDX2('c','u'):	return _("Cuba");
@@ -236,6 +241,7 @@ char *misoname(char *str) {
 /* .kw */	case NDX2('k','w'):	return _("Kuwait");
 /* .ky */	case NDX2('k','y'):	return _("Cayman Islands");
 /* .kz */	case NDX2('k','z'):	return _("Kazakhstan");
+/* .la */	case NDX2('l','a'):	return _("Laos");
 /* .lb */	case NDX2('l','b'):	return _("Lebanon");
 /* .lc */	case NDX2('l','c'):	return _("Saint Lucia");
 /* .li */	case NDX2('l','i'):	return _("Liechtenstein");
@@ -269,6 +275,7 @@ char *misoname(char *str) {
 /* .my */	case NDX2('m','y'):	return _("Malaysia");
 /* .mz */	case NDX2('m','z'):	return _("Mozambique");
 /* .na */	case NDX2('n','a'):	return _("Namibia");
+/* .name */	case NDX4('n','a','m','e'):	return _("Personal Name (.name)");
 /* .nato */	case NDX4('n','a','t','o'):	return _("Nato field");
 /* .nc */	case NDX2('n','c'):	return _("New Caledonia");
 /* .ne */	case NDX2('n','e'):	return _("Niger");
@@ -362,7 +369,7 @@ char *misoname(char *str) {
 /* .zm */	case NDX2('z','m'):	return _("Zambia");
 /* .zr */	case NDX2('z','r'):	return _("Zaire");
 /* .zw */	case NDX2('z','w'):	return _("Zimbabwe");
-		
+
 	        case NDX4('u','n','r','e'): return _("Unresolved");
 
 /* unknown domains:
