@@ -2,7 +2,7 @@
 ** Modular Logfile Analyzer
 ** Copyright 2000 Jan Kneschke <jan@kneschke.de>
 **
-** Homepage: http://www.kneschke.de/projekte/modlogan
+** Homepage: http://www.modlogan.org
 **
 
     This program is free software; you can redistribute it and/or modify
@@ -22,7 +22,7 @@
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 
 **
-** $Id: template.c,v 1.8 2002/01/02 23:51:20 ostborn Exp $
+** $Id: template.c,v 1.21 2003/04/18 18:43:48 ostborn Exp $
 */
 
 #include <libintl.h>
@@ -40,7 +40,7 @@
 
 #include "mconfig.h"
 
-//#define MTIMER_ENABLED
+/*#define MTIMER_ENABLED*/
 
 #ifdef MTIMER_ENABLED
 #include "misc.h"
@@ -51,132 +51,126 @@
 
 /**
  * get a \n terminated line from the stringbuffer
- * 
- * a string buffer is a linear string (tmpl_string) with variable str_pos 
+ *
+ * a string buffer is a linear string (tmpl_string) with variable str_pos
  * locator which is moved from \n to \n
- * 
+ *
  * The line is copied to the buffer of the tmpl_string for further processing
- * 
+ *
  * @param conf
- * @return -1 error (conf == NULL), 0 end of string, 1 new string at 
+ * @return -1 error (conf == NULL), 0 end of string, 1 new string at
  * @see tmpl_string
  */
 
 int tmpl_get_line_from_string (tmpl_string *conf) {
 	const char *s = conf->string + conf->str_pos;
 	int i = 0;
-	
+
 	if (!conf) return -1;
-	
+
 	if (!*s) return 0;
-	
+
 	for (i = 0; *(s + i) != '\n' && *(s + i) != '\0'; i++) ;
-	
+
 	if (*(s + i) == '\n') i++;
-	
+
 	if (i > conf->buf_len) {
 		free(conf->buffer);
 		conf->buf_len = i;
 		conf->buffer = malloc((conf->buf_len + 1) * sizeof(char));
 	}
-	
+
 	strncpy(conf->buffer, s, i);
 	conf->buffer[i] = '\0';
-	
+
 	conf->str_pos += i;
-	
+
 	return 1;
 }
 
 /**
  * get a \n terminated line from a file
- * 
+ *
  * the buffer element of 'conf' contains the extracted new line
- * 
+ *
  * @param conf
  * @return -1 conf is invalid (NULL), 0 end of file, 1 a new line
- * 
+ *
  */
 
 int tmpl_get_line_from_file (tmpl_file *conf) {
 	int newline = 1;
-	
+
 	if (!conf) return -1;
-	
+
 	if (!fgets(conf->buffer, conf->buf_len-1, conf->f)) {
 		newline = 0;
 	}
 
 	while (newline && conf->buffer[strlen(conf->buffer)-1] != '\n') {
 		conf->buffer = realloc(conf->buffer, (conf->buf_len+conf->buf_inc+1) * sizeof(char));
-		
+
 		if (!fgets(conf->buffer+strlen(conf->buffer), conf->buf_inc-1, conf->f)) {
 			newline = 0;
 		}
-		
+
 		conf->buf_len += conf->buf_inc;
 	}
-	
+
 	return newline;
 }
 
 /**
  * insert a key into the variable array
- * 
+ *
  */
 
 int tmpl_insert_key(tmpl_main *tmpl, const char *s, const char *def_val) {
 	int i;
-	
+
 	if (!tmpl) return -1;
-	
+
 	/* init the tmpl_keys structure at the first run */
 	if (tmpl->tmpl_keys == NULL) {
 		tmpl->tmpl_keys_size = 16;
 		tmpl->tmpl_keys_pos = 0;
 		tmpl->tmpl_keys = malloc(sizeof(tmpl_key *) * tmpl->tmpl_keys_size);
-		
+
 		for (i = 0; i < tmpl->tmpl_keys_size; i++) {
 			tmpl->tmpl_keys[i] = malloc(sizeof(tmpl_key));
-			tmpl->tmpl_keys[i]->name = NULL;
-			tmpl->tmpl_keys[i]->value = NULL;
-			tmpl->tmpl_keys[i]->defvalue = NULL;
+			memset(tmpl->tmpl_keys[i], 0, sizeof(tmpl_key));
 		}
 	}
-	
+
 	/* increase the array of keys if neccesary */
 	if (tmpl->tmpl_keys_size == tmpl->tmpl_keys_pos) {
 		tmpl->tmpl_keys_size += 16;
 		tmpl->tmpl_keys = realloc(tmpl->tmpl_keys, sizeof(tmpl_key *) * tmpl->tmpl_keys_size);
-		
+
 		for (i = tmpl->tmpl_keys_pos; i < tmpl->tmpl_keys_size; i++) {
 			tmpl->tmpl_keys[i] = malloc(sizeof(tmpl_key));
-			tmpl->tmpl_keys[i]->name = NULL;
-			tmpl->tmpl_keys[i]->value = NULL;
-			tmpl->tmpl_keys[i]->defvalue = NULL;
+			memset(tmpl->tmpl_keys[i], 0, sizeof(tmpl_key));
 		}
 	}
-	
+
 	/* search for the key */
 	for (i = 0; i < tmpl->tmpl_keys_pos; i++) {
 		if (0 == strcmp(tmpl->tmpl_keys[i]->name, s)) {
 			break;
 		}
 	}
-	
+
 	/* if key is not found insert it in the array it the last position */
 	if (i == tmpl->tmpl_keys_pos) {
-		tmpl->tmpl_keys[tmpl->tmpl_keys_pos]->name = malloc(strlen(s) + 1);
-		strcpy(tmpl->tmpl_keys[tmpl->tmpl_keys_pos]->name, s);
-		
+		tmpl->tmpl_keys[tmpl->tmpl_keys_pos]->name = strdup(s);
+
 		if (def_val) {
-			tmpl->tmpl_keys[tmpl->tmpl_keys_pos]->defvalue = malloc(strlen(def_val) + 1);
-			strcpy(tmpl->tmpl_keys[tmpl->tmpl_keys_pos]->defvalue, def_val);
+			tmpl->tmpl_keys[tmpl->tmpl_keys_pos]->defvalue = strdup(def_val);
 		}
-	
+
 		tmpl->tmpl_keys_pos++;
 	}
-	
+
 	return 0;
 }
 
@@ -184,86 +178,98 @@ int tmpl_insert_key(tmpl_main *tmpl, const char *s, const char *def_val) {
 
 /**
  * append a string to the current block
- * 
+ *
  * @param tmpl template handle
- * @param s string 
+ * @param s string
  * @return -1 hard error, 0 no error
  */
 
 int tmpl_current_block_append(tmpl_main *tmpl, const char *s) {
 	int i;
 	char *block;
-	
+	int value_len = s ? strlen(s) : 0;
+
 	if (!tmpl) return -1;
-	
+
 	/* init blocks at the first run */
 	if (tmpl->tmpl_blocks == NULL) {
 		tmpl->tmpl_blocks_size = 16;
 		tmpl->tmpl_blocks_pos = 0;
 		tmpl->tmpl_blocks = malloc(sizeof(tmpl_block *) * tmpl->tmpl_blocks_size);
-		
+
 		for (i = 0; i < tmpl->tmpl_blocks_size; i++) {
 			tmpl->tmpl_blocks[i] = malloc(sizeof(tmpl_block));
-			tmpl->tmpl_blocks[i]->name = NULL;
-			tmpl->tmpl_blocks[i]->value = NULL;
+			memset(tmpl->tmpl_blocks[i], 0, sizeof(tmpl_block));
 		}
 	}
-	
+
 	/* increase array if neccesary */
 	if (tmpl->tmpl_blocks_size == tmpl->tmpl_blocks_pos) {
 		tmpl->tmpl_blocks_size += 16;
 		tmpl->tmpl_blocks = realloc(tmpl->tmpl_blocks, sizeof(tmpl_block *) * tmpl->tmpl_blocks_size);
-		
+
 		for (i = tmpl->tmpl_blocks_pos; i < tmpl->tmpl_blocks_size; i++) {
 			tmpl->tmpl_blocks[i] = malloc(sizeof(tmpl_block));
-			tmpl->tmpl_blocks[i]->name = NULL;
-			tmpl->tmpl_blocks[i]->value = NULL;
+			memset(tmpl->tmpl_blocks[i], 0, sizeof(tmpl_block));
 		}
 	}
-	
+
 	/* if no 'current block' is set */
 	block = (tmpl->tmpl_current_block) ? tmpl->tmpl_current_block : TMPL_DEF_BLOCK;
-#if 0	
+#if 0
 	printf("appending '%s' to block '%s'\n", s, block);
-#endif	
-	
+#endif
+
 	/* search for the block and add the string 's' to the block */
 	for (i = 0; i < tmpl->tmpl_blocks_pos; i++) {
 		if (0 == strcmp(tmpl->tmpl_blocks[i]->name, block)) {
 			if (tmpl->tmpl_blocks[i]->value) {
-				tmpl->tmpl_blocks[i]->value = 
-					realloc(tmpl->tmpl_blocks[i]->value, 
-						strlen(tmpl->tmpl_blocks[i]->value) + strlen(s) + 1);
-				strcat(tmpl->tmpl_blocks[i]->value, s);
+				int required_size = tmpl->tmpl_blocks[i]->value_used + value_len + 1;
+				
+				if (tmpl->tmpl_blocks[i]->value_size <= required_size) {
+					tmpl->tmpl_blocks[i]->value_size = required_size + (required_size > 128 ? required_size : 128);
+					
+					tmpl->tmpl_blocks[i]->value =
+						realloc(tmpl->tmpl_blocks[i]->value,
+							tmpl->tmpl_blocks[i]->value_size);
+				}
+				
+				strcpy(tmpl->tmpl_blocks[i]->value + tmpl->tmpl_blocks[i]->value_used, s);
+				tmpl->tmpl_blocks[i]->value_used += value_len;
 			} else {
-				tmpl->tmpl_blocks[i]->value = malloc(strlen(s) + 1);
+				tmpl->tmpl_blocks[i]->value_size = value_len + (value_len > 128 ? value_len : 128);
+				tmpl->tmpl_blocks[i]->value = malloc(tmpl->tmpl_blocks[i]->value_size);
 				strcpy(tmpl->tmpl_blocks[i]->value, s);
+				tmpl->tmpl_blocks[i]->value_used = value_len;
 			}
+
 			break;
 		}
 	}
-	
+
 	/* if no block is found by this name create a new block and add the string */
 	if (i == tmpl->tmpl_blocks_pos) {
-		tmpl->tmpl_blocks[i]->name = malloc(strlen(block) + 1);
-		strcpy(tmpl->tmpl_blocks[i]->name, block);
+		tmpl->tmpl_blocks[i]->name = strdup(block);
 		
-		tmpl->tmpl_blocks[i]->value = malloc(strlen(s) + 1);
+		tmpl->tmpl_blocks[i]->value_size = value_len + (value_len > 128 ? value_len : 128);
+		tmpl->tmpl_blocks[i]->value = malloc(tmpl->tmpl_blocks[i]->value_size);
 		strcpy(tmpl->tmpl_blocks[i]->value, s);
-	
+		
+		tmpl->tmpl_blocks[i]->value_used = value_len;
+
 		tmpl->tmpl_blocks_pos++;
 	}
-#if 0	
-	printf("append -- '%s' -> '%s'\n", 
+#if 0
+	printf("append -- '%s' -> '%s'\n",
 	       tmpl->tmpl_blocks[i]->name,
 	       tmpl->tmpl_blocks[i]->value);
-#endif	
+#endif
 	return 0;
 }
 
 /**
  * set current block
- * 
+ *
  * @param tmpl template handle
  * @param key name of the current block
  * @return -1 hard error, 0 no error
@@ -271,168 +277,150 @@ int tmpl_current_block_append(tmpl_main *tmpl, const char *s) {
 
 int tmpl_set_current_block(tmpl_main *tmpl, const char *key) {
 	if (!tmpl) return -1;
-	
+
 	if (tmpl->tmpl_current_block) free(tmpl->tmpl_current_block);
-	
+
 	if (key) {
-		tmpl->tmpl_current_block = malloc(strlen(key) + 1);
-		strcpy(tmpl->tmpl_current_block, key);
+		tmpl->tmpl_current_block = strdup(key);
 	} else {
 		tmpl->tmpl_current_block = NULL;
 	}
-	
+
 	return 0;
 }
 
 /**
  * extract the template from a file
- * 
+ *
  * @param tmpl template handle
  * @param filename filename which should be parsed
- * 
+ *
  * @return 0 on success
  */
 
 int tmpl_load_template(tmpl_main *tmpl, const char *filename) {
-	const char *errptr;
-	int erroffset = 0;
 #define N 20 + 1
 	int ovector[3 * N], i;
-	pcre * tmpl_keys;
 	tmpl_file t;
 #define TMPL_DEPTH_MAX 16
 	char * tmpl_block_depth_names[TMPL_DEPTH_MAX];
 	int tmpl_block_depth = 0;
 	int lc = 0;
-#ifdef MTIMER_ENABLED	
+#ifdef MTIMER_ENABLED
 	static mtimer timer;
-	
+
 	MTIMER_START(timer);
 #endif
-	
+
 	if (!tmpl) return -1;
-	
+
 	if (!filename) {
 		M_DEBUG0(tmpl->debug_level, M_DEBUG_SECTION_GENERATION, M_DEBUG_LEVEL_ERRORS,
 			 "no template file specified\n");
 		return -1;
 	}
-	
+
 	if (!(t.f = fopen(filename, "r"))) {
 		M_DEBUG2(tmpl->debug_level, M_DEBUG_SECTION_GENERATION, M_DEBUG_LEVEL_ERRORS,
-			 "can't open template file '%s': %s\n", 
-			 filename, 
+			 "can't open template file '%s': %s\n",
+			 filename,
 			 strerror(errno));
 		return -1;
 	}
-	
+
 	t.buf_len = 128;
 	t.buf_inc = 128;
-	
+
 	t.buffer = malloc(sizeof(char) * t.buf_len);
-	
+
 	for (i = 0; i < TMPL_DEPTH_MAX; i++)
 		tmpl_block_depth_names[i] = NULL;
-	
-	if ((tmpl_keys = pcre_compile(
-				        /* tag                         | block */
-				      "({([A-Z][A-Z0-9_]*?)(?:=(.*?)|)}|<!-- (BEGIN|END) ([a-z][a-z0-9_]*) -->)", 
-				      0, &errptr, &erroffset, NULL)) == NULL) {
-		
-		M_DEBUG1(tmpl->debug_level, M_DEBUG_SECTION_PARSING, M_DEBUG_LEVEL_ERRORS,
-			 "rexexp compilation error at %s\n",
-			 errptr);
-		free(t.buffer);
-		return -1;
-	}
-	
+
 	while (tmpl_get_line_from_file(&t)) {
 		int offset = 0;
 		int n;
-		
+
 		int slen;
 		char *s;
-		
+
 		lc++;
 #if 0
-		printf("%s.%d: %s", __FILE__, __LINE__, t.buffer); 
-#endif		
-		while ((n = pcre_exec(tmpl_keys, NULL, t.buffer, strlen(t.buffer), offset, 0, ovector, 3 * N)) == 4 || n == 6 || n == 3) {
+		printf("%s.%d: %s", __FILE__, __LINE__, t.buffer);
+#endif
+		while ((n = pcre_exec(tmpl->match_tag_block, NULL, t.buffer, strlen(t.buffer), offset, 0, ovector, 3 * N)) == 4 || n == 6 || n == 3) {
 			slen = ovector[0] - offset;
 			s = malloc(slen + 1);
 			strncpy(s, t.buffer + offset, slen);
 			s[slen] = '\0';
-				
+
 			tmpl_current_block_append(tmpl, s);
-				
+
 			free(s);
-			
+
 			if (n == 4 || n == 3) { /* a key */
 				char *defval = NULL;
 				/* the key */
 				slen = ovector[5] - ovector[4];
-			
+
 				s = malloc(slen + 1);
 				strncpy(s, t.buffer + ovector[4], slen);
 				s[slen] = '\0';
-				
+
 				/* handle default value */
 				if (n == 4) {
 					slen = ovector[7] - ovector[6];
-			
+
 					defval = malloc(slen + 1);
 					strncpy(defval, t.buffer + ovector[6], slen);
 					defval[slen] = '\0';
 				}
-				
+
 				tmpl_insert_key(tmpl, s, defval);
 				if (defval) free(defval);
-				
+
 				tmpl_current_block_append(tmpl, "{");
 				tmpl_current_block_append(tmpl, s);
 				tmpl_current_block_append(tmpl, "}");
-				
+
 				free(s);
 			} else { /* a block */
 				int slen = ovector[11] - ovector[10];
 				char *s;
-				
+
 				s = malloc(slen + 1);
 				strncpy(s, t.buffer + ovector[10], slen);
 				s[slen] = '\0';
-				
+
 				if (*(t.buffer + ovector[8]) == 'B') {
 					char *block;
 					/* add {<blockname>} to the parent */
 					tmpl_current_block_append(tmpl, "{");
 					tmpl_current_block_append(tmpl, s);
 					tmpl_current_block_append(tmpl, "}");
-					
+
 					block = tmpl->tmpl_current_block ? tmpl->tmpl_current_block : TMPL_DEF_BLOCK;
-					
+
 					if (tmpl_block_depth >= TMPL_DEPTH_MAX) {
 						M_DEBUG2(tmpl->debug_level, M_DEBUG_SECTION_PARSING, M_DEBUG_LEVEL_ERRORS,
 							 "line %d: max. depth (%d) of blocks reached\n",
 							 lc,
 							 TMPL_DEPTH_MAX);
-						pcre_free(tmpl_keys);
 						free(t.buffer);
 						return -1;
 					}
-					
-					tmpl_block_depth_names[tmpl_block_depth] = malloc(strlen(block) + 1);
-					strcpy(tmpl_block_depth_names[tmpl_block_depth], block);
+
+					tmpl_block_depth_names[tmpl_block_depth] = strdup(block);
 					tmpl_block_depth++;
-					
+
 					/* BEGIN */
 					tmpl_set_current_block(tmpl, s);
 				} else {
 					/* END */
-					
+
 					if (tmpl_block_depth > 0) {
 						if (0 == strcmp(tmpl->tmpl_current_block, s)) {
 							tmpl_block_depth--;
-							
+
 							tmpl_set_current_block(tmpl, tmpl_block_depth_names[tmpl_block_depth]);
 							free(tmpl_block_depth_names[tmpl_block_depth]);
 							tmpl_block_depth_names[tmpl_block_depth] = NULL;
@@ -442,7 +430,6 @@ int tmpl_load_template(tmpl_main *tmpl, const char *filename) {
 								 lc,
 								 tmpl->tmpl_current_block,
 								 s);
-							pcre_free(tmpl_keys);
 							free(t.buffer);
 							return -1;
 						}
@@ -451,202 +438,181 @@ int tmpl_load_template(tmpl_main *tmpl, const char *filename) {
 							 "line %d: more END (%s) then BEGIN\n",
 							 lc,
 							 s);
-						pcre_free(tmpl_keys);
 						free(t.buffer);
 						return -1;
 					}
 				}
-			
+
 				free(s);
 			}
 			offset = ovector[1];
 		}
-		
+
 		if (n < 0 && n != PCRE_ERROR_NOMATCH) {
 			M_DEBUG1(tmpl->debug_level, M_DEBUG_SECTION_PARSING, M_DEBUG_LEVEL_ERRORS,
 				 "execution error while matching: %d\n", n);
-			pcre_free(tmpl_keys);
 			free(t.buffer);
 			return M_RECORD_HARD_ERROR;
 		}
-		
+
 		/* the rest */
 		slen = strlen(t.buffer) - offset;
 		s = malloc(slen + 1);
 		strncpy(s, t.buffer + offset, slen);
 		s[slen] = '\0';
-		
+
 		tmpl_current_block_append(tmpl, s);
-			
+
 		free(s);
 	}
-	
+
 	if (tmpl_block_depth > 0) {
 		M_DEBUG2(tmpl->debug_level, M_DEBUG_SECTION_PARSING, M_DEBUG_LEVEL_ERRORS,
 			 "line %d: missing END tag for %s\n",
 			 lc,
 			 tmpl_block_depth_names[tmpl_block_depth]);
-		pcre_free(tmpl_keys);
 		free(t.buffer);
 		return -1;
 	}
-	
+
 	fclose(t.f);
-#ifdef MTIMER_ENABLED	
+#ifdef MTIMER_ENABLED
 	MTIMER_STOP(timer);
 	MTIMER_CALC(timer);
-	
+
 	fprintf(stderr, "timer %s: %ld msec\n",
 		__FUNCTION__, timer.span );
-#endif	
-	
+#endif
+
 	free(t.buffer);
-	pcre_free(tmpl_keys);
 	return 0;
 }
 
 
 /**
  * extract the template from a string
- * 
+ *
  * @param tmpl template handle
  * @param filename filename which should be parsed
- * 
+ *
  * @return 0 on success
  */
 
 int tmpl_load_string(tmpl_main *tmpl, const char *string) {
-	const char *errptr;
-	int erroffset = 0;
 #define N 20 + 1
 	int ovector[3 * N], i;
-	pcre * tmpl_keys;
 	tmpl_string t;
 #define TMPL_DEPTH_MAX 16
 	char * tmpl_block_depth_names[TMPL_DEPTH_MAX];
 	int tmpl_block_depth = 0;
 	int lc = 0;
-#ifdef MTIMER_ENABLED	
+#ifdef MTIMER_ENABLED
 	static mtimer timer;
-	
+
 	MTIMER_START(timer);
 #endif
-	
+
 	if (!tmpl) return -1;
-	
+
 	if (!string) {
 		M_DEBUG0(tmpl->debug_level, M_DEBUG_SECTION_GENERATION, M_DEBUG_LEVEL_ERRORS,
 			 "no template string specified\n");
 		return -1;
 	}
-	
+
 	t.buf_len = 128;
-	
+
 	t.buffer = malloc(sizeof(char) * t.buf_len);
 	t.string = string;
 	t.str_pos = 0;
-	
+
 	for (i = 0; i < TMPL_DEPTH_MAX; i++)
 		tmpl_block_depth_names[i] = NULL;
-	
-	if ((tmpl_keys = pcre_compile(
-				        /* tag                         | block */
-				      "({([A-Z][A-Z0-9_]*?)(?:=(.*?)|)}|<!-- (BEGIN|END) ([a-z][a-z0-9_]*) -->)", 
-				      0, &errptr, &erroffset, NULL)) == NULL) {
-		
-		M_DEBUG1(tmpl->debug_level, M_DEBUG_SECTION_PARSING, M_DEBUG_LEVEL_ERRORS,
-			 "rexexp compilation error at %s\n",
-			 errptr);
-		free(t.buffer);
-		return -1;
-	}
-	
+
 	while (tmpl_get_line_from_string(&t)) {
 		int offset = 0;
 		int n;
-		
+
 		int slen;
 		char *s;
-		
+
 		lc++;
 #if 0
-		printf("%s.%d: %s", __FILE__, __LINE__, t.buffer); 
-#endif		
-		while ((n = pcre_exec(tmpl_keys, NULL, t.buffer, strlen(t.buffer), offset, 0, ovector, 3 * N)) == 4 || n == 6 || n == 3) {
+		printf("%s.%d: %s", __FILE__, __LINE__, t.buffer);
+#endif
+		while ((n = pcre_exec(tmpl->match_tag_block, NULL, t.buffer, strlen(t.buffer), offset, 0, ovector, 3 * N)) == 4 || n == 6 || n == 3) {
 			slen = ovector[0] - offset;
 			s = malloc(slen + 1);
 			strncpy(s, t.buffer + offset, slen);
 			s[slen] = '\0';
-				
+
 			tmpl_current_block_append(tmpl, s);
-				
+
 			free(s);
-			
+
 			if (n == 4 || n == 3) { /* a key */
 				char *defval = NULL;
 				/* the key */
 				slen = ovector[5] - ovector[4];
-			
+
 				s = malloc(slen + 1);
 				strncpy(s, t.buffer + ovector[4], slen);
 				s[slen] = '\0';
-				
+
 				/* handle default value */
 				if (n == 4) {
 					slen = ovector[7] - ovector[6];
-			
+
 					defval = malloc(slen + 1);
 					strncpy(defval, t.buffer + ovector[6], slen);
 					defval[slen] = '\0';
 				}
-				
+
 				tmpl_insert_key(tmpl, s, defval);
-				
+
 				tmpl_current_block_append(tmpl, "{");
 				tmpl_current_block_append(tmpl, s);
 				tmpl_current_block_append(tmpl, "}");
-				
+
 				free(s);
 			} else { /* a block */
 				int slen = ovector[11] - ovector[10];
 				char *s;
-				
+
 				s = malloc(slen + 1);
 				strncpy(s, t.buffer + ovector[10], slen);
 				s[slen] = '\0';
-				
+
 				if (*(t.buffer + ovector[8]) == 'B') {
 					char *block;
 					/* add {<blockname>} to the parent */
 					tmpl_current_block_append(tmpl, "{");
 					tmpl_current_block_append(tmpl, s);
 					tmpl_current_block_append(tmpl, "}");
-					
+
 					block = tmpl->tmpl_current_block ? tmpl->tmpl_current_block : TMPL_DEF_BLOCK;
-					
+
 					if (tmpl_block_depth >= TMPL_DEPTH_MAX) {
 						M_DEBUG2(tmpl->debug_level, M_DEBUG_SECTION_PARSING, M_DEBUG_LEVEL_ERRORS,
 							 "line %d: max. depth (%d) of blocks reached\n",
 							 lc,
 							 TMPL_DEPTH_MAX);
 						free(t.buffer);
-						pcre_free(tmpl_keys);
 						return -1;
 					}
-					
-					tmpl_block_depth_names[tmpl_block_depth] = malloc(strlen(block) + 1);
-					strcpy(tmpl_block_depth_names[tmpl_block_depth], block);
+
+					tmpl_block_depth_names[tmpl_block_depth] = strdup(block);
 					tmpl_block_depth++;
-					
+
 					/* BEGIN */
 					tmpl_set_current_block(tmpl, s);
 				} else {
 					/* END */
-					
+
 					if (tmpl_block_depth > 0) {
 						if (0 == strcmp(tmpl->tmpl_current_block, s)) {
 							tmpl_block_depth--;
-							
+
 							tmpl_set_current_block(tmpl, tmpl_block_depth_names[tmpl_block_depth]);
 							free(tmpl_block_depth_names[tmpl_block_depth]);
 							tmpl_block_depth_names[tmpl_block_depth] = NULL;
@@ -657,7 +623,6 @@ int tmpl_load_string(tmpl_main *tmpl, const char *string) {
 								 tmpl->tmpl_current_block,
 								 s);
 							free(t.buffer);
-							pcre_free(tmpl_keys);
 							return -1;
 						}
 					} else {
@@ -666,269 +631,264 @@ int tmpl_load_string(tmpl_main *tmpl, const char *string) {
 							 lc,
 							 s);
 						free(t.buffer);
-						pcre_free(tmpl_keys);
 						return -1;
 					}
 				}
-			
+
 				free(s);
 			}
 			offset = ovector[1];
 		}
-		
+
 		if (n < 0 && n != PCRE_ERROR_NOMATCH) {
 			M_DEBUG1(tmpl->debug_level, M_DEBUG_SECTION_PARSING, M_DEBUG_LEVEL_ERRORS,
 				 "execution error while matching: %d\n", n);
 			free(t.buffer);
-			pcre_free(tmpl_keys);
 			return M_RECORD_HARD_ERROR;
 		}
-		
+
 		/* the rest */
 		slen = strlen(t.buffer) - offset;
 		s = malloc(slen + 1);
 		strncpy(s, t.buffer + offset, slen);
 		s[slen] = '\0';
-		
+
 		tmpl_current_block_append(tmpl, s);
-			
+
 		free(s);
 	}
-	
+
 	if (tmpl_block_depth > 0) {
 		M_DEBUG2(tmpl->debug_level, M_DEBUG_SECTION_PARSING, M_DEBUG_LEVEL_ERRORS,
 			 "line %d: missing END tag for %s\n",
 			 lc,
 			 tmpl_block_depth_names[tmpl_block_depth]);
 		free(t.buffer);
-		pcre_free(tmpl_keys);
+		return -1;
+	}
+
+#ifdef MTIMER_ENABLED
+	MTIMER_STOP(timer);
+	MTIMER_CALC(timer);
+
+	fprintf(stderr, "timer %s: %ld msec\n",
+		__FUNCTION__, timer.span );
+#endif
+	free(t.buffer);
+	return 0;
+}
+
+/**
+ * transform a block of the template into the resulting string
+ *
+ * @return transformed string
+ */
+
+int tmpl_replace_block(tmpl_main *tmpl, const char *block, buffer *replaced_block) {
+	const char *errptr;
+	int erroffset = 0;
+#define N 20 + 1
+	int ovector[3 * N], n, i;
+	pcre * tmpl_keys;
+	tmpl_string t;
+	buffer *str;
+#ifdef MTIMER_ENABLED
+	static mtimer timer;
+#endif
+	
+	if (!tmpl) return -1;
+	
+#ifdef MTIMER_ENABLED
+	MTIMER_START(timer);
+#endif
+	t.buf_len = 128;
+
+	t.buffer = malloc(sizeof(char) * t.buf_len);
+
+	for (i = 0; i < tmpl->tmpl_blocks_pos; i++) {
+		if (0 == strcmp(tmpl->tmpl_blocks[i]->name, block)) {
+			break;
+		}
+	}
+
+	if (i == tmpl->tmpl_blocks_pos) {
+		fprintf(stderr, "%s.%d: block '%s' is unknown\n", 
+			__FILE__, __LINE__,
+			block);
+		return -1;
+	}
+
+	t.string = tmpl->tmpl_blocks[i]->value;
+	t.str_pos = 0;
+
+	if ((tmpl_keys = pcre_compile(
+		"{(.+?)}",
+		0, &errptr, &erroffset, NULL)) == NULL) {
+
+		fprintf(stderr, "%s.%d: rexexp compilation error at %s\n", __FILE__, __LINE__, errptr);
+		free(t.buffer);
 		return -1;
 	}
 	
-#ifdef MTIMER_ENABLED	
+	
+	str = replaced_block;
+	
+	str->used = 0;
+	
+	while (tmpl_get_line_from_string(&t)) {
+		int offset = 0;
+		int slen;
+		char *s;
+#if 0		
+		fprintf(stderr, "%s.%d: %s\n", __FILE__, __LINE__, t.buffer);
+#endif
+		while ((n = pcre_exec(tmpl_keys, NULL, t.buffer, strlen(t.buffer), offset, 0, ovector, 3 * N)) == 2) {
+			/* before the { */
+			slen = ovector[0] - offset;
+			
+			buffer_strcat_len(str, t.buffer + offset, slen);
+
+			/* the substitution */
+			slen = ovector[3] - ovector[2];
+
+			/* find the key and substitute it */
+			for (i = 0; i < tmpl->tmpl_keys_pos; i++) {
+				if (0 == strncmp(tmpl->tmpl_keys[i]->name, t.buffer + ovector[2], slen)) {
+#if 0					
+					fprintf(stderr, "%s.%d: %s\n", __FILE__, __LINE__, tmpl->tmpl_keys[i]->name);
+#endif
+					if (tmpl->tmpl_keys[i]->value) {
+						buffer_strcat(str, tmpl->tmpl_keys[i]->value);
+					} else if (tmpl->tmpl_keys[i]->defvalue) {
+						buffer_strcat(str, tmpl->tmpl_keys[i]->defvalue);
+					} else {
+#if 0
+						fprintf(stdout, "%s.%d: %s - no value\n",
+							__FILE__, __LINE__,
+							tmpl->tmpl_keys[i]->name);
+#endif
+					}
+					break;
+				}
+			}
+
+			if (i == tmpl->tmpl_keys_pos) {
+				/* ok, we have to extract the key for fprintf */
+				s = malloc(slen + 1);
+				strncpy(s, t.buffer + ovector[2], slen);
+				s[slen] = '\0';
+				M_DEBUG2(tmpl->debug_level, M_DEBUG_SECTION_PARSING, M_DEBUG_LEVEL_WARNINGS,
+					 "key '%s' not found in block '%s'\n",
+					 s, block);
+				free(s);
+			}
+
+			offset = ovector[3] + 1;
+		}
+
+		if (n < 0 && n != PCRE_ERROR_NOMATCH) {
+			M_DEBUG1(tmpl->debug_level, M_DEBUG_SECTION_PARSING, M_DEBUG_LEVEL_ERRORS,
+				 "execution error while matching: %d\n", n);
+			free(t.buffer);
+			pcre_free(tmpl_keys);
+			return -1;
+		}
+
+		/* the rest */
+		slen = strlen(t.buffer) - offset;
+
+		/* copy the rest to the end of the string */
+		buffer_strcat_len(str, t.buffer + offset, slen);
+	}
+#ifdef MTIMER_ENABLED
 	MTIMER_STOP(timer);
 	MTIMER_CALC(timer);
-	
+
 	fprintf(stderr, "timer %s: %ld msec\n",
 		__FUNCTION__, timer.span );
-#endif	
+#endif
 	free(t.buffer);
 	pcre_free(tmpl_keys);
 	return 0;
 }
 
-/**
- * transform a block of the template into the resulting string
- * 
- * @return transformed string
- */
-
-char * tmpl_replace_block(tmpl_main *tmpl, const char *block) {
-	const char *errptr;
-	int erroffset = 0;
-#define N 20 + 1
-	int ovector[3 * N], n, i;
-	pcre * tmpl_keys;
-	tmpl_string t, str;
-	int n_a = 0, w_a = 0;
-#ifdef MTIMER_ENABLED
-	static mtimer timer;
-#endif	
-	if (!tmpl) return NULL;
-#ifdef MTIMER_ENABLED	
-	MTIMER_START(timer);
-#endif	
-	t.buf_len = 128;
-	
-	t.buffer = malloc(sizeof(char) * t.buf_len);
-	
-	for (i = 0; i < tmpl->tmpl_blocks_pos; i++) {
-		if (0 == strcmp(tmpl->tmpl_blocks[i]->name, block)) {
-			break;
-		}
-	}
-	
-	if (i == tmpl->tmpl_blocks_pos) {
-		printf("block '%s' is unknown\n", block);
-		return NULL;
-	}
-	
-	t.string = tmpl->tmpl_blocks[i]->value;
-	t.str_pos = 0;
-	
-	if ((tmpl_keys = pcre_compile(
-		"{(.+?)}", 
-		0, &errptr, &erroffset, NULL)) == NULL) {
-		
-		fprintf(stderr, "%s.%d: rexexp compilation error at %s\n", __FILE__, __LINE__, errptr);
-		free(t.buffer);
-		return NULL;
-	} 
-	
-	str.buf_len = 128;
-	str.buffer = malloc(str.buf_len);
-	*str.buffer = '\0';
-	str.str_pos = 0;
-	
-	while (tmpl_get_line_from_string(&t)) {
-		int offset = 0;
-		int slen;
-		char *s;
-		while ((n = pcre_exec(tmpl_keys, NULL, t.buffer, strlen(t.buffer), offset, 0, ovector, 3 * N)) == 2) {
-			/* before the { */
-			slen = ovector[0] - offset;
-			
-			if (str.buf_len < (str.str_pos + slen + 8) /* safety bytes */ ) {
-				str.buf_len += slen + 128;
-				str.buffer = realloc(str.buffer, str.buf_len);
-				w_a++;
-			} else {
-				n_a++;
-			}
-			
-			/* copy the first element to the end of the string */
-			strncpy(str.buffer + str.str_pos, t.buffer + offset, slen);
-			str.str_pos += slen;
-			*(str.buffer + str.str_pos) = '\0';
-			
-			/* the substitution */
-			slen = ovector[3] - ovector[2];
-			
-			for (i = 0; i < tmpl->tmpl_keys_pos; i++) {
-				if (0 == strncmp(tmpl->tmpl_keys[i]->name, t.buffer + ovector[2], slen)) {
-					int len;
-					if (tmpl->tmpl_keys[i]->value) {
-						len = strlen(tmpl->tmpl_keys[i]->value);
-						if (str.buf_len < (str.str_pos + len + 8) /* safety bytes */ ) {
-							str.buf_len += len + 128;
-							str.buffer = realloc(str.buffer, str.buf_len);
-							w_a++;
-						} else {
-							n_a++;
-						}
-						
-						strcpy(str.buffer + str.str_pos, tmpl->tmpl_keys[i]->value);
-						str.str_pos += len;
-					} else if (tmpl->tmpl_keys[i]->defvalue) {
-						len = strlen(tmpl->tmpl_keys[i]->defvalue);
-						if (str.buf_len < (str.str_pos + len + 8) /* safety bytes */ ) {
-							str.buf_len += len + 128;
-							str.buffer = realloc(str.buffer, str.buf_len);
-							w_a++;
-						} else {
-							n_a++;
-						}
-						
-						strcpy(str.buffer + str.str_pos, tmpl->tmpl_keys[i]->defvalue);
-						str.str_pos += len;
-					}
-					break;
-				}
-			}
-			
-			if (i == tmpl->tmpl_keys_pos) {
-				/* ok, we have to extract the key for fprintf */
-				s = malloc(slen + 1);
-				strncpy(s, t.buffer + ovector[2], slen);
-				s[slen] = '\0';
-				M_DEBUG1(tmpl->debug_level, M_DEBUG_SECTION_PARSING, M_DEBUG_LEVEL_WARNINGS,
-					 "key %s not found\n",
-					 s);
-				free(s);
-			}
-
-			offset = ovector[3] + 1;
-		}
-		
-		if (n < 0 && n != PCRE_ERROR_NOMATCH) {
-			M_DEBUG1(tmpl->debug_level, M_DEBUG_SECTION_PARSING, M_DEBUG_LEVEL_ERRORS,
-				 "execution error while matching: %d\n", n);
-			free(t.buffer);
-			pcre_free(tmpl_keys);
-			return NULL;
-		}
-		
-		/* the rest */
-		slen = strlen(t.buffer) - offset;
-		
-		if (str.buf_len < (str.str_pos + slen + 8) /* safety bytes */ ) {
-			str.buf_len += slen + 128;
-			str.buffer = realloc(str.buffer, str.buf_len);
-			w_a++;
-		} else {
-			n_a++;
-		}
-			
-		/* copy the rest to the end of the string */
-		strncpy(str.buffer + str.str_pos, t.buffer + offset, slen);
-		str.str_pos += slen;
-		*(str.buffer + str.str_pos) = '\0';
-	}
-#ifdef MTIMER_ENABLED	
-	MTIMER_STOP(timer);
-	MTIMER_CALC(timer);
-	
-	fprintf(stderr, "timer %s: %ld msec\n",
-		__FUNCTION__, timer.span );
-#endif	
-#if 0	
-	fprintf(stderr, "timer %s: %d - %d = %d\n",
-		__FUNCTION__, w_a, n_a, w_a - n_a );
-#endif
-	free(t.buffer);
-	pcre_free(tmpl_keys);
-	return str.buffer;
-}
-
 /**
  * transform template to a string
  */
 
-char * tmpl_replace(tmpl_main *tmpl) {
-	if (!tmpl) return NULL;
-	
-	return tmpl_replace_block(tmpl, TMPL_DEF_BLOCK);
+int tmpl_replace(tmpl_main *tmpl, buffer *replaced_block) {
+	if (!tmpl) return -1;
+
+	return tmpl_replace_block(tmpl, TMPL_DEF_BLOCK, replaced_block);
 }
 
 /**
  * set a template variable to a value
- * 
+ *
  * the value is copied into the internal data store and can be changed externally
  */
 
 int tmpl_set_var (tmpl_main *tmpl, const char *key, const char *value) {
 	int i;
-#ifdef MTIMER_ENABLED	
+#ifdef MTIMER_ENABLED
 	static mtimer timer;
-	
+
 	MTIMER_START(timer);
-#endif	
+#endif
 	if (!tmpl) return -1;
 	if (!value) return -1;
-	
+
 	for (i = 0; i < tmpl->tmpl_keys_pos; i++) {
 		if (0 == strcmp(tmpl->tmpl_keys[i]->name, key)) {
-			if (tmpl->tmpl_keys[i]->value) free(tmpl->tmpl_keys[i]->value);
+			int value_len = strlen(value);
+			
+			if (tmpl->tmpl_keys[i]->value) {
+				if (value_len >= tmpl->tmpl_keys[i]->value_size) {
+					/* we have to resize */
+					if (tmpl->tmpl_keys[i]->value_size <= value_len) {
+						tmpl->tmpl_keys[i]->value_size = value_len + (value_len > 128 ? value_len : 128);
+						
+						tmpl->tmpl_keys[i]->value =
+							realloc(tmpl->tmpl_keys[i]->value,
+								tmpl->tmpl_keys[i]->value_size);
+					}
+				} else {
+					/* the old string is large enough */
+				}
+			} else {
+				/* we need some new space */
+				tmpl->tmpl_keys[i]->value_size = value_len + (value_len > 128 ? value_len : 128);
+				tmpl->tmpl_keys[i]->value = malloc(tmpl->tmpl_keys[i]->value_size);
+			}
 			
-			tmpl->tmpl_keys[i]->value = malloc(strlen(value) + 1);
 			strcpy(tmpl->tmpl_keys[i]->value, value);
-			
+			tmpl->tmpl_keys[i]->value_used = value_len;
+
 			break;
 		}
 	}
+
+	if (i == tmpl->tmpl_keys_pos) {
+#if 0
+		fprintf(stderr, "%s.%d: variable '%s' not found in block\n",
+			__FILE__, __LINE__,
+			key);
+#endif
+	}
+
 #ifdef MTIMER_ENABLED
 	MTIMER_STOP(timer);
 	MTIMER_CALC(timer);
-	
+
 	fprintf(stderr, "timer %s: %ld msec\n",
 		__FUNCTION__, timer.span );
 #endif
-	return -(i == tmpl->tmpl_keys_pos);
+	return (i == tmpl->tmpl_keys_pos) ? -1 : 0;
 }
 
 /**
  * append a value to a variable 'key'
- * 
+ *
  * @param tmpl template handle
  * @param key name of the variable
  * @param value value which shall be appended
@@ -939,34 +899,50 @@ int tmpl_append_var (tmpl_main *tmpl, const char *key, const char *value) {
 	int i;
 #ifdef MTIMER_ENABLED
 	static mtimer timer;
-	
+
 	MTIMER_START(timer);
-#endif	
+#endif
 	if (!tmpl) return -1;
-	
+
 	/* append the value to the specified variable */
 	for (i = 0; i < tmpl->tmpl_keys_pos; i++) {
 		if (0 == strcmp(tmpl->tmpl_keys[i]->name, key)) {
+			int value_len = strlen(value);
 			if (tmpl->tmpl_keys[i]->value) {
-				tmpl->tmpl_keys[i]->value = 
-					realloc(tmpl->tmpl_keys[i]->value, 
-						strlen(tmpl->tmpl_keys[i]->value) + strlen(value) + 1);
-				strcat(tmpl->tmpl_keys[i]->value, value);
+				int required_size = tmpl->tmpl_keys[i]->value_used + value_len + 1;
+				
+				/* resize if neccesary */
+				/* have to reduces some allocations 
+				 * 
+				 * 179555447  74852           0      0  template.c:951
+				 */
+				if (tmpl->tmpl_keys[i]->value_size <= required_size) {
+					tmpl->tmpl_keys[i]->value_size = required_size + (required_size > 128 ? required_size : 128);
+					
+					tmpl->tmpl_keys[i]->value =
+						realloc(tmpl->tmpl_keys[i]->value,
+							tmpl->tmpl_keys[i]->value_size);
+				}
+				
+				strcpy(tmpl->tmpl_keys[i]->value + tmpl->tmpl_keys[i]->value_used, value);
+				tmpl->tmpl_keys[i]->value_used += value_len;
 			} else {
-				tmpl->tmpl_keys[i]->value = malloc(strlen(value) + 1);
+				tmpl->tmpl_keys[i]->value_size = value_len + (value_len > 128 ? value_len : 128);
+				tmpl->tmpl_keys[i]->value = malloc(tmpl->tmpl_keys[i]->value_size);
 				strcpy(tmpl->tmpl_keys[i]->value, value);
+				tmpl->tmpl_keys[i]->value_used = value_len;
 			}
-			
+
 			break;
 		}
 	}
-#ifdef MTIMER_ENABLED	
+#ifdef MTIMER_ENABLED
 	MTIMER_STOP(timer);
 	MTIMER_CALC(timer);
-	
+
 	fprintf(stderr, "timer %s: %ld msec\n",
 		__FUNCTION__, timer.span );
-#endif	
+#endif
 	/* return -1 if variable is not found */
 	return -(i == tmpl->tmpl_keys_pos);
 }
@@ -979,25 +955,26 @@ int tmpl_clear_var (tmpl_main *tmpl, const char *key) {
 	int i;
 #ifdef MTIMER_ENABLED
 	static mtimer timer;
-	
+
 	MTIMER_START(timer);
-#endif	
+#endif
 	if (!tmpl) return -1;
-	
+
 	for (i = 0; i < tmpl->tmpl_keys_pos; i++) {
 		if (0 == strcmp(tmpl->tmpl_keys[i]->name, key)) {
 			if (tmpl->tmpl_keys[i]->value) free(tmpl->tmpl_keys[i]->value);
 			tmpl->tmpl_keys[i]->value = NULL;
+			tmpl->tmpl_keys[i]->value_size = 0;
 			break;
 		}
 	}
-#ifdef MTIMER_ENABLED	
+#ifdef MTIMER_ENABLED
 	MTIMER_STOP(timer);
 	MTIMER_CALC(timer);
-	
+
 	fprintf(stderr, "timer %s: %ld msec\n",
 		__FUNCTION__, timer.span );
-#endif	
+#endif
 	return -(i == tmpl->tmpl_keys_pos);
 }
 
@@ -1006,15 +983,15 @@ int tmpl_clear_var (tmpl_main *tmpl, const char *key) {
  */
 
 int tmpl_clear_block (tmpl_main *tmpl, const char *key) {
-	return tmpl_clear_var(tmpl, key); 
+	return tmpl_clear_var(tmpl, key);
 }
 
 /**
  * parse the current block and add the transformed block to the main block
- * 
+ *
  * Sideeffect:
  * sets the current block to the default block after processing
- * 
+ *
  */
 
 int tmpl_parse_current_block(tmpl_main *tmpl) {
@@ -1022,39 +999,35 @@ int tmpl_parse_current_block(tmpl_main *tmpl) {
 	int i;
 #ifdef MTIMER_ENABLED
 	static mtimer timer;
-	
+
 	MTIMER_START(timer);
-#endif	
+#endif
 	if (!tmpl) return -1;
-	
+
 	block = (tmpl->tmpl_current_block) ? tmpl->tmpl_current_block : TMPL_DEF_BLOCK;
-	
+
 	for (i = 0; i < tmpl->tmpl_blocks_pos; i++) {
 		if (0 == strcmp(tmpl->tmpl_blocks[i]->name, block)) {
-			char *s = tmpl_replace_block(tmpl, block);
-			
-			if (s) {
+			if (0 == tmpl_replace_block(tmpl, block, tmpl->tmp_buf)) {
 				tmpl_insert_key(tmpl, block, NULL);
-				tmpl_append_var(tmpl, block, s);
-			
-				free(s);
+				tmpl_append_var(tmpl, block, tmpl->tmp_buf->ptr);
 			}
-			
+
 			break;
 		}
 	}
-	
+
 	if (i == tmpl->tmpl_blocks_pos) {
 		M_DEBUG1(tmpl->debug_level, M_DEBUG_SECTION_PARSING, M_DEBUG_LEVEL_WARNINGS,
-			 "block %s not found\n", 
+			 "block %s not found\n",
 			 block);
 	}
-	
+
 	tmpl_set_current_block(tmpl, NULL);
-#ifdef MTIMER_ENABLED	
+#ifdef MTIMER_ENABLED
 	MTIMER_STOP(timer);
 	MTIMER_CALC(timer);
-	
+
 	fprintf(stderr, "timer %s: %ld msec\n",
 		__FUNCTION__, timer.span );
 #endif
@@ -1063,32 +1036,32 @@ int tmpl_parse_current_block(tmpl_main *tmpl) {
 
 
 /**
- * remove all keys 
- * 
+ * remove all keys
+ *
  */
 int tmpl_free_keys(tmpl_main *tmpl) {
 	int i;
 #ifdef MTIMER_ENABLED
 	static mtimer timer;
-	
+
 	MTIMER_START(timer);
-#endif	
+#endif
 	if (!tmpl) return -1;
 	if (!tmpl->tmpl_keys) return -1;
-	
+
 	for (i = 0; i < tmpl->tmpl_keys_size; i++) {
 		if (tmpl->tmpl_keys[i]->value) free(tmpl->tmpl_keys[i]->value);
 		if (tmpl->tmpl_keys[i]->defvalue) free(tmpl->tmpl_keys[i]->defvalue);
 		if (tmpl->tmpl_keys[i]->name) free(tmpl->tmpl_keys[i]->name);
 		free(tmpl->tmpl_keys[i]);
 	}
-	
+
 	free(tmpl->tmpl_keys);
 	tmpl->tmpl_keys = NULL;
-#ifdef MTIMER_ENABLED	
+#ifdef MTIMER_ENABLED
 	MTIMER_STOP(timer);
 	MTIMER_CALC(timer);
-	
+
 	fprintf(stderr, "timer %s: %ld msec\n",
 		__FUNCTION__, timer.span );
 #endif
@@ -1097,30 +1070,30 @@ int tmpl_free_keys(tmpl_main *tmpl) {
 
 
 /**
- * remove all blocks 
+ * remove all blocks
  */
 int tmpl_free_blocks(tmpl_main *tmpl) {
 	int i;
 #ifdef MTIMER_ENABLED
 	static mtimer timer;
-	
+
 	MTIMER_START(timer);
-#endif	
+#endif
 	if (!tmpl) return -1;
 	if (!tmpl->tmpl_blocks) return -1;
-	
+
 	for (i = 0; i < tmpl->tmpl_blocks_size; i++) {
 		if (tmpl->tmpl_blocks[i]->value) free(tmpl->tmpl_blocks[i]->value);
 		if (tmpl->tmpl_blocks[i]->name) free(tmpl->tmpl_blocks[i]->name);
 		free(tmpl->tmpl_blocks[i]);
 	}
-	
+
 	free(tmpl->tmpl_blocks);
 	tmpl->tmpl_blocks = NULL;
 #ifdef MTIMER_ENABLED
 	MTIMER_STOP(timer);
 	MTIMER_CALC(timer);
-	
+
 	fprintf(stderr, "timer %s: %ld msec\n",
 		__FUNCTION__, timer.span );
 #endif
@@ -1133,12 +1106,27 @@ int tmpl_free_blocks(tmpl_main *tmpl) {
 
 tmpl_main *tmpl_init() {
 	tmpl_main *tmpl;
-	
+	const char *errptr;
+	int erroffset = 0;
+
 	tmpl = malloc(sizeof(tmpl_main));
-	
+
 	/* set everything to zero */
 	memset(tmpl, 0, sizeof(tmpl_main));
 	
+	if ((tmpl->match_tag_block = pcre_compile(
+				      /* tag                         | block */
+				      "({([A-Z][A-Z0-9_]*?)(?:=(.*?)|)}|<!-- (BEGIN|END) ([a-z][a-z0-9_]*) -->)",
+				      0, &errptr, &erroffset, NULL)) == NULL) {
+		
+		M_DEBUG1(10, M_DEBUG_SECTION_PARSING, M_DEBUG_LEVEL_ERRORS,
+			 "rexexp compilation error at %s\n",
+			 errptr);
+		return NULL;
+	}
+	
+	tmpl->tmp_buf = buffer_init();
+	
 	return tmpl;
 }
 
@@ -1149,7 +1137,11 @@ int tmpl_free(tmpl_main *tmpl) {
 	tmpl_free_blocks(tmpl);
 	tmpl_free_keys(tmpl);
 	
-	free(tmpl);
+	pcre_free(tmpl->match_tag_block);
 	
+	buffer_free(tmpl->tmp_buf);
+
+	free(tmpl);
+
 	return 0;
 }
