@@ -2,7 +2,7 @@
 ** Modular Logfile Analyzer
 ** Copyright 2000 Jan Kneschke <jan@kneschke.de>
 **
-** Homepage: http://www.kneschke.de/projekte/modlogan
+** Homepage: http://www.modlogan.org
 **
 
     This program is free software; you can redistribute it and/or modify
@@ -22,7 +22,7 @@
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 
 **
-** $Id: generate.c,v 1.29 2002/01/03 12:18:16 ostborn Exp $
+** $Id: generate.c,v 1.39 2003/01/06 11:19:21 ostborn Exp $
 */
 
 #include <libintl.h>
@@ -38,6 +38,7 @@
 #include <unistd.h>
 
 #include "mconfig.h"
+#include "misc.h"
 #include "mstate.h"
 #include "mlocale.h"
 #include "mhash.h"
@@ -56,93 +57,23 @@
 #define NO_VCOUNT	16
 #define COUNTRY 	32
 
-/* tla -> Three Letter Abbreviation */
-char *get_month_string(int m, int tla) {
-	static char monthname[255];
-	
-	struct tm *tm;
-	time_t t = time(NULL);
-	
-	tm = localtime(&t);
-	
-	tm->tm_mon = m > 0 ? m - 1 : 11;
-	
-	strftime(monthname, sizeof(monthname)-1, tla ? "%b" : "%B", tm);
-	return monthname;
-}
-
-char *htmlencode(const char *s) {
-	char *p;
-	char *q = NULL;
-	int q_len = 0;
-	if (!s) return NULL;
-	
-	q_len = strlen(s) * 2 + 1;
-	q = malloc(q_len);
-
-	p = q;
-	*p = '\0';
-	
-	while (*s) {
-		switch(*s) {
-			case '<':
-				*p = '&';
-				*(++p) = 'l';
-				*(++p) = 't';
-				*(++p) = ';';
-				break;
-			case '>':
-				*p = '&';
-				*(++p) = 'g';
-				*(++p) = 't';
-				*(++p) = ';';
-				break;
-			default:
-				*p = *s;
-				break;
-		}
-		*(++p) = '\0';
-		s++;
-		
-		if (strlen(q) > (q_len - 4)) {
-			q_len += 128;
-			q = realloc(q, q_len);
-			
-			p = q + strlen(q);
-		}
-	}
-	
-	return q;
-}
-
-
 int mhash_get_value(mhash *h, const char *key) {
 	int c = 0, i;
 	if (!h) return 0;
-	
+
 	for ( i = 0; i < h->size; i++) {
-		mlist *l = h->data[i]->list;
-		while (l) {
-			if (l->data) {
-				mdata *data = l->data;
-				
-				if (!strcmp(key, data->key)) {
-					switch (data->type) {
-						case M_DATA_TYPE_COUNT:
-							c = data->data.count.count;
-							break;
-						default:
-							fprintf(stderr, "%s.%d: can't get count from a uncountable datatype (%d)\n", __FILE__, __LINE__, data->type);
-							c = 0;
-					}
-					break;
-				}
+		mlist *l;
+		
+		for (l = h->data[i]->list; l && l->data; l = l->next) {
+			mdata *data = l->data;
+			
+			if (0 == strcmp(key, data->key)) {
+				return mdata_get_count(data);
 			}
-			l = l->next;
 		}
 		if (c != 0) break;
 	}
-	
+
 	return c;
 }
 
@@ -150,52 +81,46 @@ int show_mhash (mconfig *ext_conf, FILE *f, mhash *h, int count, int opt) {
 	mlist *l, *first;
 	int i = 0;
 	config_output *conf = ext_conf->plugin_conf;
-	
+
 	if (!h) return 0;
-	
+
 	first = l = mlist_init();
-	
+
 	mhash_unfold_sorted_limited(h, l, count);
-	
+
 	while (l && (i++ < count)) {
 		mdata *data = l->data;
-		
+
 		if (data) {
-			char *enc_url = htmlencode(data->key);
-			int cut_url = strlen(enc_url) > 40;
+			char *enc_url;
+			int cut_url;
 			
+			enc_url = html_encode(data->key);
+			cut_url = strlen(enc_url) > 40;
+
 			if (cut_url) {
 				enc_url[40] = '\0';
 			}
-			
+
 			fprintf(f,"<TR>");
-			
+
 			if (opt & INDEX) {
 				fprintf(f,"<TD width=\"5%%\" align=right>%d</TD>", i);
 			}
-			
-			if (data->type == M_DATA_TYPE_COUNT) {
-				fprintf(f,"<TD width=\"15%%\" align=right>%d</TD>", data->data.count.count);
-			} else if (data->type == M_DATA_TYPE_VISITED) {
-				fprintf(f,"<TD width=\"15%%\" align=right>%d</TD>", data->data.visited.count);
-				if (!(opt & NO_VCOUNT)) {
-					fprintf(f,"<TD width=\"15%%\" align=right>%d</TD>", data->data.visited.vcount);
-				}
-			} else if (data->type == M_DATA_TYPE_BROKENLINK) {
-				fprintf(f,"<TD width=\"15%%\" align=right>%d</TD>", data->data.brokenlink.count);
+
+			fprintf(f,"<TD width=\"15%%\" align=right>%d</TD>", mdata_get_count(data));
+			if (!(opt & NO_VCOUNT)) {
+				fprintf(f,"<TD width=\"15%%\" align=right>%.0f</TD>", mdata_get_vcount(data));
 			}
-			
-			if ((opt & GROUPING) && 
-				(( data->type == M_DATA_TYPE_COUNT &&
-				data->data.count.grouped == M_DATA_STATE_GROUPED) ||
-				( data->type == M_DATA_TYPE_VISITED &&
-				data->data.visited.grouped == M_DATA_STATE_GROUPED))) {
+
+			if ((opt & GROUPING) &&
+			    mdata_is_grouped(data)) {
 				fprintf(f,"<TD bgcolor=\"%s\">%s%s</TD>", conf->col_grouping, enc_url, cut_url ? "..." : "");
 			} else {
 				if (opt & HIGHLIGHT) {
 					fprintf(f,"<TD><a href=\"%s\">%s</a>%s</TD>", data->key, enc_url, cut_url ? "..." : "");
 				} else if (opt & COUNTRY) {
-					char *c = htmlencode(misoname(data->key));
+					char *c = html_encode(misoname(data->key));
 					fprintf(f,"<td>%s</td>\n", c);
 					free(c);
 				} else {
@@ -203,13 +128,13 @@ int show_mhash (mconfig *ext_conf, FILE *f, mhash *h, int count, int opt) {
 				}
 			}
 			fprintf(f,"</TR>\n");
-			
+
 			free(enc_url);
 		}
-		
+
 		l = l->next;
 	}
-	
+
 	mlist_free(first);
 
 	return 0;
@@ -221,34 +146,30 @@ int mhash_status_unfold_sorted_limited(mhash *h, mlist *l, int count ) {
 	char __dummy__[] = "999";
 	char __dummy_2_[] = "";
 	char *max, *last = __dummy_2_;
-	
+
 	for ( j = 0; j < count; j++) {
 		data = NULL;
-		
+
 		max = __dummy__;
-		
+
 		for ( i = 0; i < h->size; i++) {
 			if (h->data[i]->list) {
 				mlist *hl;
-			
-				hl = h->data[i]->list;
-				while (hl) {
-					if (hl->data) {
-						if ( strcmp(hl->data->key, max) < 0 &&
-							strcmp(hl->data->key, last) > 0) {
-							max = hl->data->key;
-							data = hl->data;
-						}
+
+				for (hl = h->data[i]->list; hl && hl->data; hl = hl->next) {
+					if ( strcmp(hl->data->key, max) < 0 &&
+					     strcmp(hl->data->key, last) > 0) {
+						max = hl->data->key;
+						data = hl->data;
 					}
-					hl = hl->next;
 				}
 			}
 		}
-		
+
 		if (data) {
 			if (data->type == M_DATA_TYPE_COUNT) {
 				ins_data = mdata_Count_create(data->key, data->data.count.count, M_DATA_STATE_PLAIN);
-			
+
 				mlist_insert(l, ins_data);
 			} else {
 				fprintf(stderr, "%s.%d: ARGS\n", __FILE__, __LINE__);
@@ -256,29 +177,30 @@ int mhash_status_unfold_sorted_limited(mhash *h, mlist *l, int count ) {
 			last = data->key;
 		}
 	}
-	
+
 	return 0;
 }
 
 int show_status_mhash (FILE *f, mhash *h, int count) {
 	mlist *l, *first;
 	int i = 0;
-	
+
 	if (!h) return 0;
-	
+
 	first = l = mlist_init();
-	
+
 	mhash_status_unfold_sorted_limited(h, l, count);
-	
+
 	while (l && (i++ < count)) {
 		mdata *data = l->data;
-		
-		if (data)
+
+		if (data) {
 			fprintf(f,"<TR><TD width=\"15%%\" align=right>%i</TD><TD>%s - %s</TD></TR>\n", data->data.count.count, data->key, mhttpcodes(strtol(data->key, NULL, 10)));
-		
+		}
+
 		l = l->next;
 	}
-	
+
 	mlist_free(first);
 
 	return 0;
@@ -296,8 +218,8 @@ void file_start(FILE *f, mconfig *ext_conf) {
 	config_output *conf = ext_conf->plugin_conf;
 
 	fprintf(f, "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">\n\n");
-	fprintf(f, "<HTML><HEAD><TITLE>Statistics</TITLE></HEAD><BODY bgcolor=\"%s\">\n", conf->col_body);
-	
+	fprintf(f, "<HTML><HEAD><TITLE>Statistics</TITLE></HEAD><meta name=\"ROBOTS\" content=\"NOINDEX, NOARCHIVE, NOFOLLOW\">\n<BODY bgcolor=\"%s\">\n", conf->col_body);
+
 	fprintf(f, "%s %s<HR><br>\n", _("Statistics for"), conf->hostname);
 }
 
@@ -306,22 +228,21 @@ void file_end(FILE *f) {
 	fprintf(f, "<HR><a href=\"http://validator.w3.org/check/referer\"><img border=0 "\
 		"src=\"http://validator.w3.org/images/vh40\" "\
 		"alt=\"Valid HTML 4.0!\" height=31 width=88 align=\"right\"></a>");
-	fprintf(f, "Output generated by <a href=\"%s\">%s %s</a>\n", "http://www.kneschke.de/projekte/modlogan/", PACKAGE, VERSION);
+	fprintf(f, "Output generated by <a href=\"%s\">%s %s</a>\n", "http://www.modlogan.org/", PACKAGE, VERSION);
 	fprintf(f, "</BODY></HTML>\n");
 }
 
 char *table_header(int maxcount, int count, char *str) {
 	static char trans_buf[255];
-	
+
 	sprintf(trans_buf, "%i %s %i %s", (maxcount > count) || (maxcount < 0) ? count : maxcount, _("of"), count, str);
-	
+
 	return trans_buf;
-	
+
 }
 
 int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate *state, const char *subpath) {
 	unsigned int i, s_200, s_304;
-	unsigned int min, sec;
 	double d = 0;
 	FILE *f;
 	char filename[255];
@@ -331,42 +252,42 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 	config_output *conf = ext_conf->plugin_conf;
 
 	mstate_web *staweb = NULL;
-	
+
 	if (!state->ext) return -1;
-	
+
 	if (state->ext_type != M_STATE_TYPE_WEB) return -1;
-	
+
 	staweb = state->ext;
-	
+
 	if (subpath) {
 		sprintf(filename, "%s/%s/",
-			ext_conf->plugin_conf ? ext_conf->outputdir : ".",
+			ext_conf->plugin_conf ? conf->outputdir : ".",
 			subpath);
 		mkdir(filename, 0755);
-		
-		sprintf(filename, "%s/%s/m_usage_%04i%02i.html", 
-			ext_conf->outputdir ? ext_conf->outputdir : ".",
+
+		sprintf(filename, "%s/%s/m_usage_%04i%02i.html",
+			conf->outputdir ? conf->outputdir : ".",
 			subpath,
 			state->year, state->month);
 	} else {
-		sprintf(filename, "%s/m_usage_%04i%02i.html", 
-			ext_conf->outputdir ? ext_conf->outputdir : ".",
+		sprintf(filename, "%s/m_usage_%04i%02i.html",
+			conf->outputdir ? conf->outputdir : ".",
 			state->year, state->month);
 	}
-	
+
 	if (!(f = fopen(filename, "w"))) {
 		return -1;
 	}
-	
+
 	file_start(f,ext_conf);
-	
+
 	sumdat.files	= maxdat.files		= 0;
 	sumdat.xfersize	= maxdat.xfersize	= 0;
 	sumdat.hits	= maxdat.hits		= 0;
 	sumdat.hosts	= maxdat.hosts		= 0;
 	sumdat.pages	= maxdat.pages		= 0;
 	sumdat.visits	= maxdat.visits		= 0;
-	
+
 	/* count the values */
 	for ( i = 0; i < 31; i++) {
 		if (staweb->days[i].hits) last_day = i+1;
@@ -379,22 +300,22 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 
 		if (maxdat.files < staweb->days[i].files)
 			maxdat.files	= staweb->days[i].files;
-		if (maxdat.hits < staweb->days[i].hits) 
+		if (maxdat.hits < staweb->days[i].hits)
 			maxdat.hits	= staweb->days[i].hits;
-		if (maxdat.hosts < staweb->days[i].hosts) 
+		if (maxdat.hosts < staweb->days[i].hosts)
 			maxdat.hosts	= staweb->days[i].hosts;
-		if (maxdat.pages < staweb->days[i].pages) 
+		if (maxdat.pages < staweb->days[i].pages)
 			maxdat.pages	= staweb->days[i].pages;
-		if (maxdat.visits < staweb->days[i].visits) 
+		if (maxdat.visits < staweb->days[i].visits)
 			maxdat.visits	= staweb->days[i].visits;
-		if (maxdat.xfersize < staweb->days[i].xfersize) 
+		if (maxdat.xfersize < staweb->days[i].xfersize)
 			maxdat.xfersize	= staweb->days[i].xfersize;
 	}
-	
-	
-	
+
+
+
 	table_start(f, _("Summary"), 3);
-	
+
 	/* Totals */
 	fprintf(f,"<TR><TD>%s</TD><TD colspan=2 align=right>%li</TD></TR>\n", _("Total Hits"), sumdat.hits);
 	fprintf(f,"<TR><TD>%s</TD><TD colspan=2 align=right>%li</TD></TR>\n", _("Total Files"), sumdat.files);
@@ -402,16 +323,16 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 /*	fprintf(f,"<TR><TD>%s</TD><TD colspan=2 align=right>%li</TD></TR>\n", _("Total Hosts"), sumdat.hosts);*/
 	fprintf(f,"<TR><TD>%s</TD><TD colspan=2 align=right>%li</TD></TR>\n", _("Total Visits"), sumdat.visits);
 	fprintf(f,"<TR><TD>%s</TD><TD colspan=2 align=right>%.0f</TD></TR>\n", _("Transfered KBytes"), sumdat.xfersize / 1024);
-	
+
 	fprintf(f,"<TR><TH>&nbsp;</TH><TH>%s</TH><TH>%s</TH></TR>\n", _("avg"), _("max"));
-	
+
 	fprintf(f,"<TR><TD>%s</TD><TD align=right>%li</TD><TD align=right>%li</TD></TR>\n", _("Hits per Day"), sumdat.hits / last_day, maxdat.hits);
 	fprintf(f,"<TR><TD>%s</TD><TD align=right>%li</TD><TD align=right>%li</TD></TR>\n", _("Files per Day"), sumdat.files / last_day, maxdat.files);
 	fprintf(f,"<TR><TD>%s</TD><TD align=right>%li</TD><TD align=right>%li</TD></TR>\n", _("Pages per Day"), sumdat.pages / last_day, maxdat.pages);
 /*	fprintf(f,"<TR><TD>%s</TD><TD align=right>%li</TD><TD align=right>%li</TD></TR>\n", _("Hosts per Day"), sumdat.hosts / last_day, maxdat.hosts);*/
 	fprintf(f,"<TR><TD>%s</TD><TD align=right>%li</TD><TD align=right>%li</TD></TR>\n", _("Visits per Day"), sumdat.visits / last_day, maxdat.visits);
 	fprintf(f,"<TR><TD>%s</TD><TD align=right>%.0f</TD><TD align=right>%.0f</TD></TR>\n", _("Transfered Kbytes per Day"), (sumdat.xfersize / 1024) / last_day, maxdat.xfersize / 1024);
-#if 0	
+#if 0
 	if (sumdat.visits) {
 		d = (staweb->allvisittime / sumdat.visits);
 		min = d / 60;
@@ -421,7 +342,7 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 		sec = 0;
 	}
 	fprintf(f,"<TR><TD>%s</TD><TD align=right>%i:%02i %s</TD><TD align=right>%s</TD></TR>\n", _("Time per visit"), min, sec, _("min"), "---");
-	
+
 	if (sumdat.visits) {
 		d = (double)staweb->allvisitlength / sumdat.visits;
 	} else {
@@ -431,20 +352,20 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 #endif
 	s_200 = mhash_get_value(staweb->status_hash, "200");
 	s_304 = mhash_get_value(staweb->status_hash, "304");
-	
+
 	d = ((double)s_304/(s_200+s_304)) * 100;
 	fprintf(f,"<TR><TD>%s</TD><TD align=right>%.2f%%</TD><TD align=right>%s</TD></TR>\n", _("Cache Hit ratio"), d, "---");
-	
+
 	table_end(f);
-	
+
 	ref = create_pic_31_day(ext_conf, state, subpath);
-	
+
 	if (ref && strlen(ref)) {
 		fprintf(f, "%s", ref);
 	}
-	
-	
-	
+
+
+
 	table_start(f, _("Daily Statistics"), 6);
 	fprintf(f,"<TR><TH>%s</TH><TH bgcolor=\"%s\">%s</TH><TH bgcolor=\"%s\">%s</TH><TH bgcolor=\"%s\">%s</TH><TH bgcolor=\"%s\">%s</TH><TH bgcolor=\"%s\">%s</TH></TR>\n",
 		_("Day"),
@@ -461,7 +382,7 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 		);
 	for ( i = 0; i < last_day; i++) {
 		fprintf(f,"<TR><TD>%i</TD><TD align=right>%li</TD><TD align=right>%li</TD><TD align=right>%li</TD><TD align=right>%li</TD><TD align=right>%.0f</TD></TR>\n",
-			i+1, 
+			i+1,
 			staweb->days[i].hits,
 			staweb->days[i].files,
 			staweb->days[i].pages,
@@ -470,13 +391,13 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 			);
 	}
 	table_end(f);
-	
+
 	ref = create_pic_24_hour(ext_conf, state, subpath);
-	
+
 	if (ref && strlen(ref)) {
 		fprintf(f, "%s", ref);
 	}
-	
+
 	table_start(f, _("Hourly Statistics"), 6);
 	fprintf(f,"<TR><TH>%s</TH><TH bgcolor=\"%s\">%s</TH><TH bgcolor=\"%s\">%s</TH><TH bgcolor=\"%s\">%s</TH><TH bgcolor=\"%s\">%s</TH><TH bgcolor=\"%s\">%s</TH></TR>\n",
 		_("Hour"),
@@ -493,7 +414,7 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 		);
 	for ( i = 0; i < 24; i++) {
 		fprintf(f,"<TR><TD>%i</TD><TD align=right>%li</TD><TD align=right>%li</TD><TD align=right>%li</TD><TD align=right>%li</TD><TD align=right>%.0f</TD></TR>\n",
-			i, 
+			i,
 			staweb->hours[i].hits,
 			staweb->hours[i].files,
 			staweb->hours[i].pages,
@@ -502,7 +423,7 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 			);
 	}
 	table_end(f);
-	
+
 	if (conf->max_req_urls > 0) {
 		table_start(f, table_header(conf->max_req_urls, mhash_count(staweb->req_url_hash),_("Requested URL's")), 3);
 		fprintf(f,"<TR><TH>#</TH><TH bgcolor=\"%s\">%s</TH><TH>%s</TH></TR>\n",
@@ -513,7 +434,7 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 		show_mhash(ext_conf, f,staweb->req_url_hash, conf->max_req_urls, HIGHLIGHT | GROUPING | INDEX | NO_VCOUNT);
 		table_end(f);
 	}
-	
+
 	if (conf->max_os > 0) {
 		table_start(f, table_header(conf->max_os, mhash_count(staweb->os_hash), _("Used Operating Systems")), 4);
 		fprintf(f,"<TR><TH>#</TH><TH bgcolor=\"%s\">%s</TH><TH bgcolor=\"%s\">%s</TH><TH>%s</TH></TR>\n",
@@ -526,7 +447,7 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 		show_mhash(ext_conf, f,staweb->os_hash, conf->max_os, GROUPING | VISITS | INDEX);
 		table_end(f);
 	}
-	
+
 	if (conf->max_hosts > 0) {
 		table_start(f, table_header(conf->max_hosts, mhash_count(staweb->host_hash),_("Hosts")), 4);
 		fprintf(f,"<TR><TH>#</TH><TH bgcolor=\"%s\">%s</TH><TH bgcolor=\"%s\">%s</TH><TH>%s</TH></TR>\n",
@@ -539,7 +460,7 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 		show_mhash(ext_conf, f,staweb->host_hash, conf->max_hosts, GROUPING | INDEX);
 		table_end(f);
 	}
-	
+
 	if (conf->max_ref_urls > 0) {
 		table_start(f, table_header(conf->max_ref_urls, mhash_count(staweb->ref_url_hash),_("Referring URL's")), 3);
 		fprintf(f,"<TR><TH>#</TH><TH bgcolor=\"%s\">%s</TH><TH>%s</TH></TR>\n",
@@ -561,7 +482,7 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 		show_mhash(ext_conf, f,staweb->entry_pages, conf->max_entry_pages, HIGHLIGHT | INDEX);
 		table_end(f);
 	}
-	
+
 	if (conf->max_exit_pages > 0) {
 		table_start(f, table_header(conf->max_exit_pages, mhash_count(staweb->exit_pages),_("Exit Pages")), 3);
 		fprintf(f,"<TR><TH>#</TH><TH bgcolor=\"%s\">%s</TH><TH>%s</TH></TR>\n",
@@ -583,7 +504,7 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 		show_mhash(ext_conf, f,staweb->indexed_pages, conf->max_indexed_pages,HIGHLIGHT | INDEX);
 		table_end(f);
 	}
-	
+
 	if (conf->max_ua > 0) {
 		table_start(f, table_header(conf->max_ua, mhash_count(staweb->ua_hash),_("Used Browsers")), 4);
 		fprintf(f,"<TR><TH>#</TH><TH bgcolor=\"%s\">%s</TH><TH bgcolor=\"%s\">%s</TH><TH>%s</TH></TR>\n",
@@ -596,7 +517,7 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 		show_mhash(ext_conf, f,staweb->ua_hash, conf->max_ua, GROUPING | VISITS | INDEX);
 		table_end(f);
 	}
-	
+
 	if (conf->max_req_prot > 0) {
 		table_start(f, table_header(conf->max_req_prot, mhash_count(staweb->req_prot_hash),_("Used Request Protocol")), 3);
 		fprintf(f,"<TR><TH>#</TH><TH bgcolor=\"%s\">%s</TH><TH>%s</TH></TR>\n",
@@ -607,7 +528,7 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 		show_mhash(ext_conf, f,staweb->req_prot_hash, conf->max_req_prot,0 | INDEX);
 		table_end(f);
 	}
-	
+
 	if (conf->max_req_meth > 0) {
 		table_start(f, table_header(conf->max_req_meth, mhash_count(staweb->req_meth_hash),_("Used Request Method")), 3);
 		fprintf(f,"<TR><TH>#</TH><TH bgcolor=\"%s\">%s</TH><TH>%s</TH></TR>\n",
@@ -618,7 +539,7 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 		show_mhash(ext_conf, f,staweb->req_meth_hash,conf->max_req_meth,0 | INDEX);
 		table_end(f);
 	}
-	
+
 	if (conf->max_status_codes > 0) {
 		table_start(f, table_header(conf->max_status_codes, mhash_count(staweb->status_hash),_("Status Code")), 2);
 		fprintf(f,"<TR><TH bgcolor=\"%s\">%s</TH><TH>%s</TH></TR>\n",
@@ -629,7 +550,7 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 		show_status_mhash(f,staweb->status_hash, conf->max_status_codes);
 		table_end(f);
 	}
-	
+
 	if (conf->max_robots > 0) {
 		table_start(f, table_header(conf->max_robots, mhash_count(staweb->robots),_("Robots")), 3);
 		fprintf(f,"<TR><TH>#</TH><TH bgcolor=\"%s\">%s</TH><TH>%s</TH></TR>\n",
@@ -640,7 +561,7 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 		show_mhash(ext_conf, f,staweb->robots, conf->max_robots,0 | INDEX);
 		table_end(f);
 	}
-	
+
 	if (conf->max_bookmarks > 0) {
 		table_start(f, table_header(conf->max_bookmarks, mhash_count(staweb->bookmarks),_("Bookmarked Pages")), 3);
 		fprintf(f,"<TR><TH>#</TH><TH bgcolor=\"%s\">%s</TH><TH>%s</TH></TR>\n",
@@ -651,7 +572,7 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 		show_mhash(ext_conf, f,staweb->bookmarks,conf->max_bookmarks,HIGHLIGHT | INDEX);
 		table_end(f);
 	}
-	
+
 	if (conf->max_broken_links > 0) {
 		table_start(f, table_header(conf->max_broken_links, mhash_count(staweb->status_missing_file),_("Missing File / Broken Link")), 3);
 		fprintf(f,"<TR><TH>#</TH><TH bgcolor=\"%s\">%s</TH><TH>%s</TH></TR>\n",
@@ -662,13 +583,13 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 		show_mhash(ext_conf, f, staweb->status_missing_file, conf->max_broken_links, GROUPING | INDEX);
 		table_end(f);
 	}
-	
+
 	if (conf->max_internal_errors > 0) {
 		table_start(f, table_header(conf->max_internal_errors, mhash_count(staweb->status_internal_error),_("Internal Errors")), 2);
 		show_mhash(ext_conf, f,staweb->status_internal_error,conf->max_internal_errors,0);
 		table_end(f);
 	}
-	
+
 	if (conf->max_search_strings > 0) {
 		table_start(f, table_header(conf->max_search_strings, mhash_count(staweb->searchstring),_("SearchStrings")), 3);
 		fprintf(f,"<TR><TH>#</TH><TH bgcolor=\"%s\">%s</TH><TH>%s</TH></TR>\n",
@@ -679,7 +600,7 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 		show_mhash(ext_conf, f,staweb->searchstring, conf->max_search_strings,0 | INDEX);
 		table_end(f);
 	}
-	
+
 	if (conf->max_search_engines > 0) {
 		table_start(f, table_header(conf->max_search_engines, mhash_count(staweb->searchsite),_("SearchEngines")), 3);
 		fprintf(f,"<TR><TH>#</TH><TH bgcolor=\"%s\">%s</TH><TH>%s</TH></TR>\n",
@@ -690,10 +611,10 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 		show_mhash(ext_conf, f,staweb->searchsite, conf->max_search_engines, HIGHLIGHT | GROUPING | INDEX);
 		table_end(f);
 	}
-	
+
 	if (conf->max_countries > 0) {
 		ref = create_pic_countries(ext_conf, state, subpath);
-	
+
 		if (ref && strlen(ref)) {
 			fprintf(f, "%s", ref);
 		}
@@ -708,11 +629,11 @@ int mplugins_output_webalizer_generate_monthly_output(mconfig *ext_conf, mstate
 		show_mhash(ext_conf, f, staweb->country_hash, conf->max_countries, VISITS | INDEX | COUNTRY);
 		table_end(f);
 	}
-	
+
 	file_end(f);
-	
+
 	fclose(f);
-	
+
 	return 0;
 }
 
@@ -722,26 +643,26 @@ int mplugins_output_webalizer_generate_history_output(mconfig *ext_conf, mlist *
 	char filename[255];
 	char *ref;
 	config_output *conf = ext_conf->plugin_conf;
-	
-	sprintf(filename, "%s%s%s/index.html", 
-		ext_conf->outputdir ? ext_conf->outputdir : ".",
+
+	sprintf(filename, "%s%s%s/index.html",
+		conf->outputdir ? conf->outputdir : ".",
 		subpath ? "/" : "",
 		subpath ? subpath : "" );
-	
+
 	if (!(f = fopen(filename, "w"))) {
 		return -1;
 	}
-	
+
 	file_start(f,ext_conf);
-	
+
 	ref = create_pic_12_month(ext_conf, history, subpath);
-	
+
 	if (ref && strlen(ref)) {
 		fprintf(f, "%s", ref);
 	}
-	
+
 	table_start(f, _("History"), 11);
-	
+
 	fprintf(f,"<TR><TH>&nbsp;</TH><TH colspan=5>%s</TH><TH colspan=5>%s</TH></TR>",
 		_("Average/day"),
 		_("Totals")
@@ -776,15 +697,15 @@ int mplugins_output_webalizer_generate_history_output(mconfig *ext_conf, mlist *
 		conf->col_xfer,
 		_("KBytes")
 		);
-	
+
 	/* go to the last element */
 	while (l->next) l = l->next;
-	
+
 	while (l) {
 		mdata *data = l->data;
-		
+
 		if (!data) break;
-		
+
 		fprintf(f,"<TR><TD><A HREF=\"m_usage_%04i%02i.html\">%s %04i</A></TD>" \
 			"<TD align=\"right\">%li</TD><TD align=\"right\">%li</TD>" \
 			"<TD align=\"right\">%li</TD><TD align=\"right\">%li</TD>" \
@@ -793,31 +714,31 @@ int mplugins_output_webalizer_generate_history_output(mconfig *ext_conf, mlist *
 			"<TD align=\"right\">%li</TD><TD align=\"right\">%li</TD>" \
 			"<TD align=\"right\">%.0f</TD>" \
 			"</TR>\n",
-			data->data.webhist.year,
-			data->data.webhist.month,
-			get_month_string(data->data.webhist.month,1),
-			data->data.webhist.year,
-			
-			data->data.webhist.hits / data->data.webhist.days_passed,
-			data->data.webhist.files / data->data.webhist.days_passed,
-			data->data.webhist.pages/ data->data.webhist.days_passed,
-			data->data.webhist.visits/ data->data.webhist.days_passed,
-			data->data.webhist.xfersize / 1024 / data->data.webhist.days_passed,
-			data->data.webhist.hits,
-			data->data.webhist.files,
-			data->data.webhist.pages,
-			data->data.webhist.visits,
-			data->data.webhist.xfersize / 1024
-			
+			data->data.hist->year,
+			data->data.hist->month,
+			get_month_string(data->data.hist->month,1),
+			data->data.hist->year,
+
+			data->data.hist->data.web.hits / data->data.hist->days_passed,
+			data->data.hist->data.web.files / data->data.hist->days_passed,
+			data->data.hist->data.web.pages/ data->data.hist->days_passed,
+			data->data.hist->data.web.visits/ data->data.hist->days_passed,
+			data->data.hist->data.web.xfersize / 1024 / data->data.hist->days_passed,
+			data->data.hist->data.web.hits,
+			data->data.hist->data.web.files,
+			data->data.hist->data.web.pages,
+			data->data.hist->data.web.visits,
+			data->data.hist->data.web.xfersize / 1024
+
 			);
 		l = l->prev;
 	}
-	
+
 	table_end(f);
-	
+
 	file_end(f);
-	
+
 	fclose(f);
-	
+
 	return 0;
 }
