@@ -2,7 +2,7 @@
 ** Modular Logfile Analyzer
 ** Copyright 2000 Jan Kneschke <jan@kneschke.de>
 **
-** Homepage: http://www.kneschke.de/projekte/modlogan
+** Homepage: http://www.modlogan.org
 **
 
     This program is free software; you can redistribute it and/or modify
@@ -22,13 +22,11 @@
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 
 **
-** $Id: generate.h,v 1.4 2000/06/16 22:32:08 jk Exp $
+** $Id: generate.h,v 1.6 2002/09/26 13:10:29 ostborn Exp $
 */
 #ifndef _GENERATE_H_
 #define _GENERATE_H_
 
 #include "plugin_config.h"
 
-char *get_month_string(int m, int tla);
-
-#endif
\ No newline at end of file
+#endif
